/*
 * Standard C-library includes.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

/*
 * Unix includes.
 */
#include <sys/time.h>
#include <sys/types.h>

/*
 * Local includes.
 */
#include "ccbclientlink.h"

/*
 * A dynamically allocated object of the following type is used to
 * encapsulate all of the resources used by this demonstration
 * program.
 */
typedef struct {
  CCBClientLink *link;   /* The communications interface of the client */
  CCBConfig *cnf;        /* The configuration of the next scan */
  long command_counter;  /* The number of commands sent so far */
  unsigned long scan;    /* The numeric ID to give the current scan */
} CCBDemoClient;

/*
 * The following are the function prototypes of the constructor and
 * destructor functions used to create and delete CCBDemoClient
 * objects.
 */
static CCBDemoClient *new_CCBDemoClient(const char *host);
static CCBDemoClient *del_CCBDemoClient(CCBDemoClient *cdc);

/*
 * Provide a function prototype of the function which implements the
 * event loop of this program.
 */
static int demo_client_loop(CCBDemoClient *cdc);

/*
 * The following are function prototypes of the callback functions used
 * to respond to events reported by the library.
 */
static CCB_LOG_MSG_FN(demo_log_msg_fn);
static CCB_CMD_ERROR_FN(demo_cmd_error_fn);
static CCB_STATUS_REPLY_FN(demo_status_reply_fn);
static CCB_INTEG_MSG_FN(demo_integ_msg_fn);
static CCB_MONITOR_MSG_FN(demo_monitor_msg_fn);

/*.......................................................................
 * This is a simple demonstration program used to test the client side of
 * the CCB communications library.
 *
 * Usage:
 *  demo_client <ccb_server_address>
 */
int main(int argc, char *argv[])
{
  CCBDemoClient *cdc;  /* The resources of the CCB client */
  int waserr = 0;      /* The completion status */
  int i;
/*
 * The only argument should specify the address of the computer
 * on which the CCB server is running.
 */
  if(argc != 2) {
    fprintf(stderr, "Usage:  demo_client <ccb_server_address>\n");
    return 1;
  };
/*
 * Create the resources of the client and attempt to connect to
 * the specified CCB server.
 */
  cdc = new_CCBDemoClient(argv[1]);
  if(!cdc)
    return 1;
/*
 * Change the phase-switch configuration.
 */
  waserr = waserr || ccb_set_phase_switch_cnf(cdc->cnf, 1,2,3,4);
/*
 * Change the calibration-diode configuration.
 */
  {
    enum {NCAL=7};
    short diode_states[NCAL];
    unsigned long diode_times[NCAL];
    for(i=0; i<NCAL; i++) {
      diode_states[i] = (i%2 ? CCB_CAL_DIODE_A : 0) |
	                (i%4 ? CCB_CAL_DIODE_B : 0);
      diode_times[i] = 10 * i;
    };
    waserr = waserr || ccb_set_cal_diode_cnf(cdc->cnf, NCAL, 1, diode_states,
					     diode_times);
  };
/*
 * Change the timing configuration.
 */
  waserr = waserr || ccb_set_timing_cnf(cdc->cnf, 250, 20, 10, 1000, 500, 4000);
/*
 * Queue a reset command.
 */
  waserr = waserr || ccb_queue_reset_cmd(cdc->link, ++cdc->command_counter);
/*
 * Queue a stop-scan command.
 */
  waserr = waserr || ccb_queue_stop_scan_cmd(cdc->link, ++cdc->command_counter,
					     cdc->cnf, ++cdc->scan);
/*
 * Queue a start-scan command to start in 10 seconds from now.
 */
  if(!waserr) {
    CCBTimeStamp now;
    if(ccb_get_timestamp(&now))
      waserr = 1;
    else {
      waserr = ccb_queue_start_scan_cmd(cdc->link,
					++cdc->command_counter,
					cdc->cnf, ++cdc->scan, now.mjd,
					now.sec + 10);
    };
  };
/*
 * Queue a standby command.
 */
  waserr = waserr || ccb_queue_standby_cmd(cdc->link, ++cdc->command_counter,
					   1);
/*
 * Queue an awaken command.
 */
  waserr = waserr || ccb_queue_awaken_cmd(cdc->link, ++cdc->command_counter);
/*
 * Queue a ping command.
 */
  waserr = waserr || ccb_queue_ping_cmd(cdc->link, ++cdc->command_counter);
/*
 * Queue a check-status command.
 */
  waserr = waserr || ccb_queue_status_request_cmd(cdc->link,
						  ++cdc->command_counter);
/*
 * Run the client event loop.
 */
  waserr = waserr || demo_client_loop(cdc);
/*
 * Shutdown the client and release its resources.
 */
  cdc = del_CCBDemoClient(cdc);
  return waserr;
}

/*.......................................................................
 * Create a new CCBDemoClient object. This contains all of the resources
 * needed by the demonstration CCB client.
 *
 * Input:
 *  host             char *  The address of the computer running the
 *                           CCB server.
 * Output:
 *  return  CCBDemoClient *  The new object, or NULL on error.
 */
static CCBDemoClient *new_CCBDemoClient(const char *host)
{
  CCBDemoClient *cdc;  /* The object to be returned */
  char errmsg[100];    /* An error message buffer */
/*
 * Allocate the container.
 */
  cdc = malloc(sizeof(CCBDemoClient));
  if(!cdc) {
    fprintf(stderr, "new_CCBDemoClient: Insufficient memory.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBDemoClient().
 */
  cdc->link = NULL;
  cdc->cnf = NULL;
  cdc->command_counter = 0;
  cdc->scan = 0;
/*
 * Create the communications interface of the client.
 */
  cdc->link = new_CCBClientLink(sizeof(errmsg), errmsg);
  if(!cdc->link)
    return del_CCBDemoClient(cdc);
/*
 * Allocate the object in which configuration updates are assembled.
 */
  cdc->cnf = new_CCBConfig();
  if(!cdc->cnf)
    return del_CCBDemoClient(cdc);
/*
 * Install the log-message delivery callback.
 */
  if(ccb_log_msg_callback(cdc->link, demo_log_msg_fn, cdc))
    return del_CCBDemoClient(cdc);
/*
 * Install the integration-message delivery callback.
 */
  if(ccb_integ_msg_callback(cdc->link, demo_integ_msg_fn, cdc))
    return del_CCBDemoClient(cdc);
/*
 * Install the monitor-data message delivery callback.
 */
  if(ccb_monitor_msg_callback(cdc->link, demo_monitor_msg_fn, cdc))
    return del_CCBDemoClient(cdc);
/*
 * Install the command-error callback.
 */
  if(ccb_cmd_error_callback(cdc->link, demo_cmd_error_fn, cdc))
    return del_CCBDemoClient(cdc);
/*
 * Install the function which receives replies to previously sent check-status
 * commands.
 */
  if(ccb_status_reply_callback(cdc->link, demo_status_reply_fn, cdc))
    return del_CCBDemoClient(cdc);
/*
 * Attempt to connect to the specified CCB server, using non-blocking I/O.
 */
  if(ccb_client_connect(cdc->link, host, 1))
    return del_CCBDemoClient(cdc);
  return cdc;
}

/*.......................................................................
 * Delete a CCBDemoClient object.
 *
 * Input:
 *  cdc     CCBDemoClient *  The object to be deleted.
 * Output:
 *  return  CCBDemoClient *  The deleted object (always NULL).
 */
static CCBDemoClient *del_CCBDemoClient(CCBDemoClient *cdc)
{
  if(cdc) {
/*
 * Delete the communications interface.
 */
    cdc->link = del_CCBClientLink(cdc->link);
/*
 * Delete the configuration object.
 */
    cdc->cnf = del_CCBConfig(cdc->cnf);
/*
 * Delete the container of the object.
 */
    free(cdc);
  };
  return NULL;
}

/*.......................................................................
 * This is the I/O-driven event loop that is used to watch for and react
 * to communications with the remote CCB server.
 *
 * Input:
 *  cdc    CCBDemoClient *  The resources of the client.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int demo_client_loop(CCBDemoClient *cdc)
{
/*
 * The event loop runs until the program is either killed, or an error
 * causes a return from this function.
 */
  while(1) {
    fd_set rfds;  /* The set of file-descriptors to watch for readability */
    fd_set wfds;  /* The set of file-descriptors to watch for writability */
    int nready;   /* The number of file descriptors ready for I/O */
    int maxfd;    /* The maximum of the file descriptors in 'rfds' and */
                  /*  'wfds'. */
/*
 * Construct the sets of file descriptors to be watched from scratch.
 */
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    maxfd = 0;
/*
 * Add the file-descriptors that the library wants us to watch, to rfds
 * and wfds.
 */
    if(ccb_client_select_args(cdc->link, &rfds, &wfds, &maxfd))
      return 1;
/*
 * Wait indefinitely for the specified I/O events.
 */
    nready = select(maxfd+1, &rfds, &wfds, NULL, NULL);
/*
 * Error?
 */
    if(nready < 1) {
      if(errno != EINTR) {
	fprintf(stderr, "select() error (%s).\n", strerror(errno));
	return 1;
      };
/*
 * At least one file-descriptor is ready for I/O.
 */
    } else {
/*
 * Perform the types of I/O indicated by select().
 */
      if(ccb_client_communicate(cdc->link,
		ccb_client_selected_io(cdc->link, &rfds, &wfds)))
	return 1;
    };
  };
}

/*.......................................................................
 * This function is called whenever a log message is received.
 */
static CCB_LOG_MSG_FN(demo_log_msg_fn)
{
  unsigned h,m,s;
/*
 * Convert the time-of-day in *ts to hours minutes and seconds.
 */
  ccb_hms_of_timestamp(*ts, &h, &m, &s);
/*
 * For the sake of example, simply write the message to stdout.
 */
  (void) fprintf(stdout, "LOG UT=%lu:%02uh%02um%02us id=%lu: %s\n",
		 ts->mjd, h, m, s, id, msg);
  return 0;
}

/*.......................................................................
 * The following function reports the contents of integrations received
 * from the manager.
 */
static CCB_INTEG_MSG_FN(demo_integ_msg_fn)
{
  unsigned h,m,s;  /* Hours, minutes, seconds */
  int i;
/*
 * Convert the time-of-day in *ts to hours minutes and seconds.
 */
  ccb_hms_of_timestamp(*ts, &h, &m, &s);
/*
 * Display the contents of the message.
 */
  (void) fprintf(stdout,
     "INTEG UT=%lu:%02uh%02um%02us scan=%lu number=%lu\n"
     "      values={", ts->mjd, h, m, s, scan, number);
  for(i=0; i<nvalues; i++)
    printf("%lu%s", values[i], i<nvalues-1 ? ", " : "}\n");
  return 0;
}

/*.......................................................................
 * The following function reports the contents of monitoring messages
 * received from the manager.
 */
static CCB_MONITOR_MSG_FN(demo_monitor_msg_fn)
{
  unsigned h,m,s;  /* Hours, minutes, seconds */
  int i;
/*
 * Convert the time-of-day in *ts to hours minutes and seconds.
 */
  ccb_hms_of_timestamp(*ts, &h, &m, &s);
/*
 * Display the contents of the message.
 */
  (void) fprintf(stdout,
     "MONITOR UT=%lu:%02uh%02um%02us scan=%lu number=%lu\n"
     "      values={", ts->mjd, h, m, s, scan, number);
  for(i=0; i<nvalues; i++)
    printf("%lu%s", values[i], i<nvalues-1 ? ", " : "}\n");
  return 0;
}

/*.......................................................................
 * This function is used to report command completion errors.
 */
static CCB_CMD_ERROR_FN(demo_cmd_error_fn)
{
  fprintf(stderr, "Command %ld failed.\n", id);
  return 0;
}

/*.......................................................................
 * This function is called whenever a reply to a previously sent
 * check-status command is received over the control link.
 */
static CCB_STATUS_REPLY_FN(demo_status_reply_fn)
{
  fprintf(stdout, "demo: Received a check-status reply (status=%lu).\n",
	  status);
  return 0;
}

