#ifndef ccbclientlink_h
#define ccbclientlink_h

/*-----------------------------------------------------------------------*
 *      This is the public header of the client portion of the CCB       *
 *                      communications library.                          *
 *-----------------------------------------------------------------------*/

/*
 * Facilitate the use of the following functions and datatypes in C++ code.
 */
#ifdef __cplusplus
#define EXTERNC extern "C"
extern "C" {
#else
#define EXTERNC
#endif

/*
 * Get datatypes that are common to both the server and client sides of
 * the CCB communications library.
 */
#include "ccbcommon.h"

/*
 * Get the declarations of system datatypes that are used in the
 * function prototypes in this file.
 */
#include <sys/types.h>   /* fd_set */
#include <poll.h>        /* struct pollfd */

/*
 * The CCBClientLink datatype holds the dynamically-allocated resources
 * that facilitate manager communications with the CCB server. The
 * contents of the structure are private to the library, so this
 * public header file only includes an opaque declaration.
 */
typedef struct CCBClientLink CCBClientLink;

/*
 * Return a CCBClientLink communications object for the manager to use to
 * communicate with the CCB server on a specified host.
 */
CCBClientLink *new_CCBClientLink(size_t errdim, char *errmsg);

/*
 * Disconnect from the CCB server associated with a given CCBClientLink
 * object, and return the resources of the object to the system.
 */
CCBClientLink *del_CCBClientLink(CCBClientLink *cl);

/*
 * Initiate connections to the telemetry and control ports of a given
 * CCB server.
 */
int ccb_client_connect(CCBClientLink *cl, const char *host, int nonblocking);

/*
 * Terminate the CCB-server connection associated with a given CCBClientLink
 * object.
 */
void ccb_client_disconnect(CCBClientLink *cl);

/*
 * Toggle whether non-blocking socket I/O is used for communications
 * with a specified CCB server.
 */
int ccb_client_non_blocking_io(CCBClientLink *cl, int on);

/*
 * Add to given select() file-descriptor sets according to the I/O
 * currently being awaited by *cl.
 */
int ccb_client_select_args(CCBClientLink *cl, fd_set *rfds, fd_set *wfds,
			   int *maxfd);

/*
 * Fill in two poll() file-descriptor event structures according to
 * the I/O that is currently being awaited by *cl.
 */
int ccb_client_poll_args(CCBClientLink *cl, int size, struct pollfd *fds,
			 int *nfds);

/*
 * Convert the contents of select() file-descriptor sets to an
 * I/O mask of the form expected by ccb_client_communicate().
 */
unsigned ccb_client_selected_io(CCBClientLink *cl, const fd_set *rfds,
			      const fd_set *wfds);

/*
 * Convert the contents of poll() file-descriptor sets to an I/O mask of
 * the form expected by ccb_client_communicate().
 */
unsigned ccb_client_polled_io(CCBClientLink *cl, const struct pollfd *fds,
			      int nfds);

/*
 * Enumerate the possible types of CCB socket I/O, each type of
 * I/O being the enumerated value of a single bit.
 */
typedef enum {
  CCB_CNTRL_READ = 1,   /* Read pending on control socket */
  CCB_CNTRL_WRITE = 2,  /* Write pending on control socket */
  CCB_TELEM_READ = 4,  /* Read pending on telemetry socket */
  CCB_TELEM_WRITE = 8  /* Write pending on telemetry socket */
} CCBClientIOStatus;

/*
 * Perform the types of socket I/O listed in the bitwise union of
 * CCBClientIOStatus values in the io argument.
 */
int ccb_client_communicate(CCBClientLink *cl, unsigned io);

/*.......................................................................
 * Command-error callback functions can be declared using the following
 * macro, whose arguments and return value are defined as follows.
 *
 * Input:
 *  cl    CCBClientLink *   The CCB client communications resource object.
 *  data           void *   Any application callback data that was
 *                          registered along with the callback function.
 *  id             long     The identifier of the command that caused
 *                          the error. This is the ID that was specified
 *                          by the manager when the command was originally
 *                          queued.
 *  status CCBCmdStatus     A general indication of the nature of the
 *                          error condition.
 * Output:
 *  return          int     0 - Don't abort ccb_client_communicate().
 *                          1 - Cause ccb_client_communicate() to return
 *                              non-zero.
 */
#define CCB_CMD_ERROR_FN(fn) EXTERNC int (fn)(CCBClientLink *cl, void *data, \
				              long id, CCBCmdStatus status)
typedef CCB_CMD_ERROR_FN(CCBCmdErrorFn);

/*
 * Register a command-error callback function with a given CCB client
 * connection.
 */
int ccb_cmd_error_callback(CCBClientLink *cl, CCBCmdErrorFn *fn, void *data);

/*
 * Queue a start-scan command to be sent to a given CCB server.
 */
int ccb_queue_start_scan_cmd(CCBClientLink *cl, long id, const CCBConfig *cnf,
                             unsigned long scan, unsigned long mjd,
			     unsigned long tod);

/*
 * Queue a stop-scan command to be sent to a given CCB server.
 */
int ccb_queue_stop_scan_cmd(CCBClientLink *cl, long id, const CCBConfig *cnf,
			    unsigned long scan);

/*
 * Queue a reset command to be sent to a given CCB server.
 */
int ccb_queue_reset_cmd(CCBClientLink *cl, long id);

/*
 * Queue a standby command to be sent to a given CCB server.
 */
int ccb_queue_standby_cmd(CCBClientLink *cl, long id,
                          unsigned short stream_mask);

/*
 * Queue an awaken command to be sent to a given CCB server.
 */
int ccb_queue_awaken_cmd(CCBClientLink *cl, long id);

/*
 * Queue an ping command to be sent to a given CCB server.
 */
int ccb_queue_ping_cmd(CCBClientLink *cl, long id);

/*
 * Return the bitwise union of the links over which replies to the
 * last ping command have ben received. If all is well, CCB_ALL_LINKS
 * is returned.
 */
unsigned ccb_ping_echos(CCBClientLink *cl);

/*
 * Queue a check-status command to be sent to a given CCB server.
 */
int ccb_queue_status_request_cmd(CCBClientLink *cl, long id);

/*
 * Queue a shutdown command to be sent to a given CCB server.
 */
int ccb_queue_shutdown_cmd(CCBClientLink *cl, long id);

/*
 * Queue a reboot command to be sent to a given CCB server.
 */
int ccb_queue_reboot_cmd(CCBClientLink *cl, long id);

/*
 * Queue a monitoring-control command to be sent to a given CCB server.
 */
int ccb_queue_monitor_cmd(CCBClientLink *cl, long id, unsigned short period);

/*
 * Queue a telemetry-stream control command to be sent to a given CCB server.
 */
int ccb_queue_telemetry_cmd(CCBClientLink *cl, long id, unsigned short streams);

/*
 * Queue a log-control command to be sent to a given CCB server.
 */
int ccb_queue_logger_cmd(CCBClientLink *cl, long id, unsigned long period);

/*
 * Control-link replies to check-status commands are delivered to the
 * manager using callback functions declared as follows.
 *
 * Input:
 *  cl     CCBClientLink *  The CCB client communications resource object.
 *  data            void *  Any application callback data that was
 *                          registered along with the callback function.
 *  status unsigned long    The status of the CCB, expressed as a bitwise
 *                          union of CCBGeneralStatus enumerators.
 * Output:
 *  return          int     0 - OK.
 *                          1 - Cause ccb_client_communicate() to return
 *                              non-zero.
 */
#define CCB_STATUS_REPLY_FN(fn) EXTERNC int (fn)(CCBClientLink *cl, \
                                                      void *data, \
                                                      unsigned long status)

typedef CCB_STATUS_REPLY_FN(CCBStatusReplyFn);

/*
 * Register the callback function to be called whenever a given CCB client
 * object receives a control-status reply message.
 */
int ccb_status_reply_callback(CCBClientLink *cl, CCBStatusReplyFn *fn,
				   void *data);

/*
 * Telemetry monitor data messages are delivered to the manager via
 * callback functions declared as follows.
 *
 * Input:
 *  cl      CCBClientLink *  The CCB client communications resource object.
 *  data             void *  Any application callback data that was
 *                           registered along with the callback function.
 *  ts       CCBTimeStamp *  The originating time of the message.
 *  scan    unsigned long    The sequential number of the parent scan.
 *  number  unsigned long    The sequential monitor frame index.
 *  values  unsigned long *  The array of 'nvalues' monitor values.
 *  nvalues      unsigned    The number of values in values[].
 * Output:
 *  return          int     0 - OK.
 *                          1 - Cause ccb_client_communicate() to return
 *                              non-zero.
 */
#define CCB_MONITOR_MSG_FN(fn) EXTERNC int (fn)(CCBClientLink *cl, \
						void *data, \
						const CCBTimeStamp *ts,    \
                                                unsigned long scan, \
						unsigned long number,        \
						const unsigned long *values, \
						unsigned nvalues)

typedef CCB_MONITOR_MSG_FN(CCBMonitorMsgFn);

/*
 * Register the callback function to be called whenever a given CCB client
 * object receives a monitor data message.
 */
int ccb_monitor_msg_callback(CCBClientLink *cl, CCBMonitorMsgFn *fn,
			     void *data);

/*
 * Telemetry integrated-data messages are delivered to the manager via
 * callback functions declared as follows.
 *
 * Input:
 *  cl      CCBClientLink *  The CCB client communications resource object.
 *  data             void *  Any application callback data that was
 *                           registered along with the callback function.
 *  ts       CCBTimeStamp *  The originating time of the message.
 *  scan    unsigned long    The sequential scan number.
 *  number  unsigned long    The sequential integration number within the scan.
 *  values  unsigned long *  The array of 'nvalues' integrated values.
 *  nvalues      unsigned    The number of values in values[].
 * Output:
 *  return            int    0 - OK.
 *                           1 - Cause ccb_client_communicate() to return
 *                               non-zero.
 */
#define CCB_INTEG_MSG_FN(fn) EXTERNC int (fn)(CCBClientLink *cl, \
					      void *data, \
					      const CCBTimeStamp *ts, \
                                              unsigned long scan, \
					      unsigned long number, \
					      const unsigned long *values,\
					      unsigned nvalues)

typedef CCB_INTEG_MSG_FN(CCBIntegMsgFn);

/*
 * Register the callback function to be called whenever a given CCB client
 * object receives an integrated-data message.
 */
int ccb_integ_msg_callback(CCBClientLink *cl, CCBIntegMsgFn *fn,
			   void *data);


/*
 * Telemetry log messages are delivered to the manager via
 * callback functions declared as follows.
 *
 * Input:
 *  cl      CCBClientLink *  The CCB client communications resource object.
 *  data             void *  Any application callback data that was
 *                           registered along with the callback function.
 *  ts       CCBTimeStamp *  The originating time of the message.
 *  msg        const char *  The text of the log message.
 *  id      unsigned long    The numeric identifier of the originating
 *                           statement within the CCB software.
 *  level     CCBLogLevel    The significance of the message.
 * Output:
 *  return            int    0 - OK.
 *                           1 - Cause ccb_client_communicate() to return
 *                               non-zero.
 */
#define CCB_LOG_MSG_FN(fn) EXTERNC int (fn)(CCBClientLink *cl, void *data,\
				            const CCBTimeStamp *ts,  \
					    const char *msg, \
					    unsigned long id, \
					    CCBLogLevel level)

typedef CCB_LOG_MSG_FN(CCBLogMsgFn);

/*
 * Register the callback function to be called whenever a given CCB client
 * object receives a log message.
 */
int ccb_log_msg_callback(CCBClientLink *cl, CCBLogMsgFn *fn, void *data);

/*.......................................................................
 * Optional callback functions report changes in CCB socket file
 * descriptors and the events that they are awaiting, to third party
 * event loops where direct access to select() or poll() is not
 * available to the application.
 *
 * Input:
 *  cl       CCBClientLink *  The CCB client communications resource object.
 *  data              void *  Any application callback data that was
 *                            registered along with the callback function.
 *  cntrl_sock         int    The socket file-descriptor currently
 *                            associated with the control link, or -1 if
 *                            there currently is no control link.
 *  telem_sock         int    The socket file-descriptor currently
 *                            associated with the telemetry link, or -1 if
 *                            there currently is no telemetry link.
 *  io            unsigned    The set of events to watch for, expressed
                              as a bitwise union of CCBClientIOStatus
 *                            enumerators.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
#define CCB_CLIENT_SOCKETS_FN(fn) EXTERNC int (fn)(CCBClientLink *cl, \
		  void *data, int cntrl_sock, int telem_sock, unsigned io)

typedef CCB_CLIENT_SOCKETS_FN(CCBClientSocketsFn);

/*
 * Register the callback function to be called to report changes in
 * either of the CCB client sockets, including the events that they
 * should be watched for.
 */
int ccb_client_sockets_callback(CCBClientLink *cl, CCBClientSocketsFn *fn,
				void *data);


#ifdef __cplusplus
}
#endif

#endif
