/* Standard C includes */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stddef.h>

/* POSIX includes */

#include <pthread.h>
#include <unistd.h>

/* Socket includes */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* Local includes */

#include "ccbtelemetryclient.h"
#include "ccbclientpriv.h"
#include "netbuf.h"
#include "ccbnetobj.h"
#include "ccbconstants.h"
#include "ccbqueue.h"
#include "ccbtelemetrymsg.h"
#include "tcpip.h"
#include "ccbprint.h"

/*-----------------------------------------------------------------------*
 * The functions in this file handle the client end of the telemetry     *
 * link to a remote CCB server.						 *
 *									 *
 * Conventions:								 *
 *  All external functions are named starting with the prefix ccb_tc_.	 *
 *  These functions must check their arguments, and surround all	 *
 *  accesses to their CCBTelemetryClient argument with calls to		 *
 *  tc_lock_client() and tc_unlock_client(), to give them and any	 *
 *  internal functions that they call, exclusive access to the contents	 *
 *  of said object.							 *
 *									 *
 *  All internal functions are named starting with the prefix tc_.	 *
 *  These functions all assume that their caller has acquired		 *
 *  exclusive access to the CCBTelemetryClient object that they are	 *
 *  passed, and are careful to always reacquire exclusive access	 *
 *  before returning to the caller, if they ever relinquish it.		 *
 *									 *
 *  To avoid non-portable recursive locking of the guard mutex,		 *
 *  internal functions temporarily relinquish exclusive access to	 *
 *  their CCBTelemetryClient object before calling manager callback	 *
 *  functions. Since ccb_log_client_msg() calls a callback function	 *
 *  of the parent CCBClientLink, it is essential to call			 *
 *  tc_error_while_locked(), instead of the above function when		 *
 *  reporting an error message while holding exclusive access to	 *
 *  a CCBTelemetryClient object. This function relinquishes exclusive	 *
 *  access for the duration of the call to ccb_log_client_msg().	 *
 *									 *
 *  Functions that don't use their return values for anything else,	 *
 *  return an integer which is 1 if an error occurred, or 0 if the	 *
 *  function was successful.						 *
 *									 *
 *  Error conditions are reported once, as close as possible to the	 *
 *  site of the error. Higher level functions, on receiving an error	 *
 *  return from an internal function, assume that the error has been	 *
 *  reported by that function, or by a function that it called. Since	 *
 *  low-level external utility functions don't report log errors, it	 *
 *  is up to the function in this module that calls one of these	 *
 *  functions to report the error.					 *
 *-----------------------------------------------------------------------*/

/*
 * Collect all callback functions and callback data.
 */
typedef struct {
  CCBMonitorMsgFn *monitor_msg_fn;            /* The monitor-data callback */
  void *monitor_msg_data;                     /* monitor_msg_fn() */
                                              /*  callback data. */
  CCBIntegMsgFn *integ_msg_fn;                /* The integrated-data */
                                              /*  callback. */
  void *integ_msg_data;                       /* integ_msg_fn() */
                                              /*  callback data. */
  CCBLogMsgFn *log_msg_fn;                    /* The log-message callback. */
  void *log_msg_data;                         /* log_msg_fn() callback data. */
} CCBTelemetryClientCallbacks;

/*
 * The following constant is designed to be assigned to any
 * CCBTelemetryClientCallbacks structure to set all of its contained
 * function pointers and data pointers to their respective versions of
 * nul. This takes advantage of the fact that the C standard states
 * that statically allocated structures are initialized as though all
 * of their members were individually assigned 0 (NB this isn't the
 * same as setting all bytes to zero when pointers are involved).  Not
 * only does this tremendously reduce the number of statements needed
 * to initialize one of these structures in the CCBTelemetryClient
 * constructor, but it is also safer, because it makes it impossible
 * to forget to initialize a newly added callback and its data
 * pointer.
 */
static const CCBTelemetryClientCallbacks init_client_callbacks;

/*
 * Enumerate the states that the telemetry link goes through.
 */
typedef enum {
  TC_LINK_DOWN,          /* The telemetry link isn't connected */
  TC_LINK_CONNECTING,    /* The telemetry link is being established */
  TC_LINK_UP,            /* The telemetry link is connected and usable */
  TC_LINK_DISCONNECTING  /* The telemetry link is being disconnected */
} TCLinkState;

/*
 * Define the object that manages the CCB server end of the telemetry link.
 */
struct CCBTelemetryClient {
  pthread_mutex_t guard;          /* The mutex which telemetrys access to the */
                                  /*  other members of this object. */
  int guard_ready;                /* True after 'guard' has been allocated */
  pthread_cond_t herald;          /* The condition variable that signals */
                                  /*  changes in I/O status to readers and */
                                  /*  writers, when using blocking I/O. */
  int herald_ready;               /* True if 'herald' has been initialized */
  int read_locked;                /* True while a thread has been granted */
                                  /*  exclusive telemetry-socket read-access */
  int blocking;                   /* True when using blocking socket I/O */
  TCLinkState link_state;         /* The internal status of the telemetry */
                                  /*  link. */
  CCBTelemetryClientCallbacks cb; /* Callback functions */
  CCBClientLink *client;              /* The parent CCB client */
  NetReadStr *nrs;                /* The network input stream */
  int log_recursion_count;        /* The number of times that */
                                  /* ccb_log_client_msg() is being */
                                  /* recursively invoked. */
};

static int tc_socket_fd(CCBTelemetryClient *tc);
static int tc_unlock_client(CCBTelemetryClient *tc);
static int tc_lock_client(CCBTelemetryClient *tc);
static int tc_send_telemetry(CCBTelemetryClient *tc);
static int tc_read_telemetry(CCBTelemetryClient *tc);
static int tc_deliver_telemetry(CCBTelemetryClient *tc,
				CCBTelemetryMessage *telem);
static int tc_connect_link(CCBTelemetryClient *tc, const char *host,
			   int nonblocking);
static void tc_disconnect_link(CCBTelemetryClient *tc);
static int tc_connection_completed(CCBTelemetryClient *tc, int sock);
static int tc_activate_link(CCBTelemetryClient *tc);
static int tc_error_while_locked(CCBTelemetryClient *tc, CCBLogLevel level,
				 unsigned long id, const char *fmt, ...);

/*.......................................................................
 * Create a new CCBTelemetryClient object.
 *
 * Input:
 *  client           CCBClientLink *  The parent CCB client.
 * Output:
 *  return  CCBTelemetryClient *  The new object, or NULL on error.
 */
CCBTelemetryClient *new_CCBTelemetryClient(CCBClientLink *client)
{
  CCBTelemetryClient *tc;  /* The object to be returned */
  char errmsg[200];        /* An error message */
  int status;              /* The return value of a POSIX function */
/*
 * Check the arguments.
 */
  if(!client) {
    ccb_log_client_msg(NULL, CCB_ERROR, CCB_LOGID(132),
		       "new_CCBTelemetryClient: NULL argument(s).\n");
    return NULL;
  };
/*
 * Check that the table of telemetry message descriptions is.
 */
  if(ccb_check_net_table(&ccb_telemetry_msg_table, errmsg,
			 sizeof(errmsg))) {
    ccb_log_client_msg(NULL, CCB_ERROR, CCB_LOGID(133),
		       "Telemetry message table: %s\n", errmsg);
    return NULL;
  };
/*
 * Allocate the container.
 */
  tc = malloc(sizeof(CCBTelemetryClient));
  if(!tc) {
    ccb_log_client_msg(NULL, CCB_ERROR, CCB_LOGID(134),
		"new_CCBTelemetryClient: Insufficient memory.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBTelemetryClient().
 */
  tc->guard_ready = 0;
  tc->herald_ready = 0;
  tc->read_locked = 0;
  tc->blocking = 1;
  tc->link_state = TC_LINK_DOWN;
  tc->cb = init_client_callbacks;
  tc->client = client;
  tc->nrs = NULL;
  tc->log_recursion_count = 0;
/*
 * Allocate the mutex which is used to sequence accesses to the new object.
 */
  status = pthread_mutex_init(&tc->guard, NULL);
  if(status) {
    ccb_log_client_msg(client, CCB_ERROR, CCB_LOGID(135),
		       "new_CCBTelemetryClient: pthread_mutex_init -> %s\n",
		       strerror(status));
    return del_CCBTelemetryClient(tc);
  };
  tc->guard_ready = 1;
/*
 * Initialize the condition variable used to signal changes in I/O status.
 */
  status = pthread_cond_init(&tc->herald, NULL);
  if(status) {
    ccb_log_client_msg(client, CCB_ERROR, CCB_LOGID(255),
	"new_CCBTelemetryClient: pthread_cond_init -> %s\n", strerror(status));
    return del_CCBTelemetryClient(tc);
  };
  tc->herald_ready = 1;
/*
 * Allocate the socket input-iterator.
 */
  tc->nrs = new_NetReadStr(-1, ccb_net_max_obj_size(&ccb_telemetry_msg_table));
  if(!tc->nrs) {
    ccb_log_client_msg(client, CCB_ERROR, CCB_LOGID(136),
	"new_CCBTelemetryClient: Insufficient memory for input stream.\n");
    return del_CCBTelemetryClient(tc);
  };
  return tc;
}

/*.......................................................................
 * Disconnect and delete a CCBTelemetryClient object.
 *
 * Input:
 *  tc     CCBTelemetryClient *  The object to be deleted.
 * Output:
 *  return CCBTelemetryClient *  The deleted object (always NULL).
 */
CCBTelemetryClient *del_CCBTelemetryClient(CCBTelemetryClient *tc)
{
  if(tc) {
/*
 * Terminate any extant connection. Note that we can't safely call
 * tc_disconnect_link() here, since that function uses members that
 * might not always be allocated before this destructor function is
 * called.
 */
    int fd = tc_socket_fd(tc);
    if(fd >= 0)
      close(fd);
/*
 * Delete the mutual exclusion guard.
 */
    if(tc->guard_ready) {
      pthread_mutex_destroy(&tc->guard);
      tc->guard_ready = 0;
    };
/*
 * Delete the condition variable.
 */
    if(tc->herald_ready) {
      pthread_cond_destroy(&tc->herald);
      tc->herald_ready = 0;
    };
/*
 * Delete the I/O iterator.
 */
    tc->nrs = del_NetReadStr(tc->nrs);
/*
 * Finally delete the container.
 */
    free(tc);
  };
  return NULL;
}

/*.......................................................................
 * Query the current status of the telemetry client.
 *
 * Input:
 *  tc       CCBTelemetryClient *  The telemetry client.
 * Output:
 *  return CCBTelemClientStatus    The current status of the
 *                                 telemetry client.
 */
CCBTelemClientStatus ccb_tc_status(CCBTelemetryClient *tc)
{
/*
 * Assume that the client is down, until proven otherwise.
 */
  int status = CCB_TELEM_CLIENT_DOWN;
/*
 * Get exclusive access to 'tc' while we examine its state.
 */
  if(tc && tc_lock_client(tc)==0) {
/*
 * Derive the status from the link status.
 */
    switch(tc->link_state) {
    case TC_LINK_CONNECTING:
      status = CCB_TELEM_CLIENT_PEND;
      break;
    case TC_LINK_UP:
      status = CCB_TELEM_CLIENT_ACTIVE;
      break;
    case TC_LINK_DOWN:
    case TC_LINK_DISCONNECTING:
      status = CCB_TELEM_CLIENT_DOWN;
      break;
    };
/*
 * Relinquish exclusive access to 'tc'.
 */
    tc_unlock_client(tc);
  };
  return status;
}

/*.......................................................................
 * Respond to the reported writability of the telemetry socket.  Since
 * telemetry clients never send data to CCB servers over the telemetry
 * link, this can only mean the completion of a non-blocking
 * connection.
 *
 * Input:
 *  tc   CCBTelemetryClient *  The telemetry client.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_tc_send_telemetry(CCBTelemetryClient *tc)
{
  int status;              /* The completion status of tc_send_telemetry() */
/*
 * Check the arguments.
 */
  if(!tc) {
    ccb_log_client_msg(tc ? tc->client : NULL, CCB_ERROR, CCB_LOGID(137),
		       "ccb_tc_send_telemetry: NULL argument(s).\n");
    return 1;
  };
/*
 * Obtain exclusive access to 'tc'.
 */
  if(tc_lock_client(tc))
    return 1;
/*
 * Respond to the writability of the socket.
 */
  status = tc_send_telemetry(tc);
/*
 * Relinquish exclusive access to 'tc'.
 */
  tc_unlock_client(tc);
  return status;
}

/*.......................................................................
 * Respond to the reported writability of the telemetry socket.  Since
 * telemetry clients never send data to CCB servers over the telemetry
 * link, this can only mean the completion of a non-blocking connection.
 *
 * Note that the caller must acquire exclusive access to 'tc' before
 * calling this function.
 *
 * Input:
 *  tc   CCBTelemetryClient *  The telemetry client.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
static int tc_send_telemetry(CCBTelemetryClient *tc)
{
/*
 * We can't write anything if the link is either down or being disconnected.
 */
  if(tc->link_state==TC_LINK_DOWN || tc->link_state==TC_LINK_DISCONNECTING) {
    return tc_error_while_locked(tc, CCB_ERROR, CCB_LOGID(138),
	 "Rejected request to send CCB commands while not connected.\n");
/*
 * If a non-blocking connection is being performed, check whether it
 * succeded.
 */
  } else if(tc->link_state == TC_LINK_CONNECTING &&
	    !tc_connection_completed(tc, ccb_tc_socket(tc))) {
    return 1;
  };
/*
 * There is nothing to be written.
 */
  return 0;
}

/*.......................................................................
 * Register the callback function to be called whenever a given CCB client
 * object receives a monitor-data telemetry message.
 *
 * Input:
 *  tc      CCBTelemetryClient *  The telemetry client.
 *  fn         CCBMonitorMsgFn *  The function to call whenever monitor
 *                                data are received over the telemetry
 *                                link.
 *                             
 *  data                  void *  Anonymous data to pass to fn().
 * Output:
 *  return                 int    0 - OK.
 *                                1 - Error.
 */
int ccb_tc_monitor_msg_callback(CCBTelemetryClient *tc,
				CCBMonitorMsgFn *fn, void *data)
{
  if(!tc || !fn) {
    ccb_log_client_msg(tc ? tc->client : NULL, CCB_ERROR, CCB_LOGID(139),
	"ccb_tc_monitor_msg_callback: NULL argument(s).\n");    
    return 1;
  };
/*
 * Temporarily acquire exclusive access to 'tc' while installing
 * the callback.
 */
  if(tc_lock_client(tc))
    return 1;
  tc->cb.monitor_msg_fn = fn;
  tc->cb.monitor_msg_data = data;
  tc_unlock_client(tc);
  return 0;
}

/*.......................................................................
 * Register the callback function to be called whenever a given CCB client
 * object receives an integrated-data telemetry message.
 *
 * Input:
 *  tc      CCBTelemetryClient *  The telemetry client.
 *  fn           CCBIntegMsgFn *  The function to call whenever integrated
 *                                data is received over the telemetry
 *                                link.
 *                             
 *  data                  void *  Anonymous data to pass to fn().
 * Output:
 *  return                 int    0 - OK.
 *                                1 - Error.
 */
int ccb_tc_integ_msg_callback(CCBTelemetryClient *tc,
			      CCBIntegMsgFn *fn, void *data)
{
  if(!tc || !fn) {
    ccb_log_client_msg(tc ? tc->client : NULL, CCB_ERROR, CCB_LOGID(140),
	"ccb_tc_integ_msg_callback: NULL argument(s).\n");    
    return 1;
  };
/*
 * Temporarily acquire exclusive access to 'tc' while installing
 * the callback.
 */
  if(tc_lock_client(tc))
    return 1;
  tc->cb.integ_msg_fn = fn;
  tc->cb.integ_msg_data = data;
  tc_unlock_client(tc);
  return 0;
}

/*.......................................................................
 * Register the callback function to be called whenever a given CCB
 * client receives a log message.
 *
 * Input:
 *  tc      CCBTelemetryClient *  The telemetry client.
 *  fn             CCBLogMsgFn *  The function to call whenever a log
 *                                message is received over the telemetry
 *                                link.
 *                             
 *  data                  void *  Anonymous data to pass to fn().
 * Output:
 *  return                 int    0 - OK.
 *                                1 - Error.
 */
int ccb_tc_log_msg_callback(CCBTelemetryClient *tc, CCBLogMsgFn *fn,
			    void *data)
{
  if(!tc || !fn) {
    ccb_log_client_msg(tc ? tc->client : NULL, CCB_ERROR, CCB_LOGID(141),
	"ccb_tc_log_msg_callback: NULL argument(s).\n");    
    return 1;
  };
/*
 * Temporarily acquire exclusive access to 'tc' while installing
 * the callback.
 */
  if(tc_lock_client(tc))
    return 1;
  tc->cb.log_msg_fn = fn;
  tc->cb.log_msg_data = data;
  tc_unlock_client(tc);
  return 0;
}

/*.......................................................................
 * Read up to one telemetry message from the CCB client telemetry socket,
 * and when a complete telemetry-message has been read, deliver it to the
 * manager by calling the appropriate callback function.
 *
 * Input:
 *  tc    CCBTelemetryClient *  The telemetry client.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
int ccb_tc_read_telemetry(CCBTelemetryClient *tc)
{
  int status;    /* The completion status of tc_read_telemetry() */
/*
 * Check the arguments.
 */
  if(!tc) {
    ccb_log_client_msg(tc ? tc->client : NULL, CCB_ERROR, CCB_LOGID(143),
		       "ccb_tc_read_telemetry: NULL argument(s).\n");
    return 1;
  };
/*
 * Obtain exclusive access to 'tc'.
 */
  if(tc_lock_client(tc))
    return 1;
/*
 * Read from the telemetry socket.
 */
  status = tc_read_telemetry(tc);
/*
 * Relinquish exclusive access to 'tc'.
 */
  tc_unlock_client(tc);
  return status;
}

/*.......................................................................
 * Read as many messages from the CCB client telemetry socket as
 * possible, and deliver complete telemetry-messages to the manager by
 * calling the appropriate callback function. In blocking I/O mode
 * this function doesn't return until the telemetry connection is lost.
 * Otherwise it returns as soon as doing otherwise would cause it to
 * wait.
 *
 * Note that the caller must acquire exclusive access to 'tc' before
 * calling this function.
 *
 * Input:
 *  tc    CCBTelemetryClient *  The telemetry client.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
static int tc_read_telemetry(CCBTelemetryClient *tc)
{
  CCBTelemetryMessage msg; /* A generic telemetry-message structure */
  int read_status;         /* The return status of nrs_read_msg() */
  int status;              /* The return status of pthread functions */
/*
 * Read as many messages as possible.
 */
  do {
/*
 * If blocking I/O has been selected, wait for exclusive read access
 * to the socket when the link is up.
 */
    while(tc->blocking && tc->read_locked) {
      status = pthread_cond_wait(&tc->herald, &tc->guard);
      if(status) {
	ccb_log_client_msg(tc->client, CCB_ERROR, CCB_LOGID(256),
			   "tc_read_telemetry: pthread_cond_wait -> %s\n",
			   strerror(status));
	return 1;
      };
    };
/*
 * If another thread is currently reading from the telemetry socket,
 * defer to it.
 */
    if(tc->read_locked)
      return 0;
/*
 * We can't read anything if the link is either down, or being disconnected.
 */
    if(tc->link_state==TC_LINK_DOWN || tc->link_state==TC_LINK_DISCONNECTING) {
      return tc_error_while_locked(tc, CCB_ERROR, CCB_LOGID(144),
	 "Rejected request to read CCB telemetry while not connected.\n");
/*
 * If a non-blocking connection is being performed, check whether it
 * succeded.
 */
    } else if(tc->link_state == TC_LINK_CONNECTING &&
	      !tc_connection_completed(tc, ccb_tc_socket(tc))) {
      return 1;
    };
/*
 * Before reading, set the flag that indicates that we have exclusive
 * read access to the telemetry socket, then give up the mutex so that
 * other threads can output messages while we potentially wait for the
 * read operation to complete.
 */
    tc->read_locked = 1;
    tc_unlock_client(tc);
/*
 * Attempt to read more data from the telemetry socket.
 */
    read_status = nrs_read_msg(tc->nrs);
/*
 * Reacquire exclusive access to 'tc' now that we are no longer waiting
 * for the read operation.
 */
    if(tc_lock_client(tc))
      return 1;
/*
 * Relinquish exclusive read access to the telemetry socket.  Note that
 * no reader will actually get a chance to see that we have done this
 * until this function returns, and the caller relinquishes access to
 * 'tc'.
 */
    tc->read_locked = 0;
    status = pthread_cond_broadcast(&tc->herald);
    if(status) {
      ccb_log_client_msg(tc->client, CCB_ERROR, CCB_LOGID(257),
			 "tc_read_telemetry: pthread_cond_broadcast -> %s\n",
			 strerror(status));
      return 1;
    };
/*
 * Check the completion status of the read operation.
 */
    switch(read_status) {
    case NET_READ_SIZE:
    case NET_READ_DATA:   /* Input blocked before a message could be read */
      if(!tc->blocking)
	return 0;
      break;
    case NET_READ_DONE:   /* A new message has now been completely read */
/*
 * Decode the message into a message structure.
 */
      if(ccb_net_to_obj(&ccb_telemetry_msg_table, nrs_get_NetBuf(tc->nrs),
			&msg.head.base)) {
	return tc_error_while_locked(tc, CCB_ERROR, CCB_LOGID(145),
		  "tc_read_telemetry: ccb_net_to_obj() error (%s)",
		  strerror(errno));
      };
/*
 * Invoke the appropriate message-delivery callback.
 */
      if(tc_deliver_telemetry(tc, &msg))
	return 1;
      break;
    case NET_READ_CLOSED: /* There is no telemetry socket */
      return tc_error_while_locked(tc, CCB_FATAL, CCB_LOGID(146),
        "The CCB server has dropped the telemetry-connection.\n");
    break;
    case NET_READ_ERROR:  /* I/O error */
    default:
      return tc_error_while_locked(tc, CCB_FATAL, CCB_LOGID(147),
  "An I/O error was encountered while reading a CCB telemetry message.\n");
      break;
    };
/*
 * Keep reading messages unless the link has been lost (or something else
 * in the above body of this loop causes this function to return, such as
 * an error, or I/O blocking when in non-blocking mode).
 */
  } while(tc->link_state != TC_LINK_UP);
  return 0;
}

/*.......................................................................
 * Invoke the appropriate callback to deliver a newly received telemetry
 * message to the manager.
 *
 * Input:
 *  tc      CCBTelemetryClient *  The telemetry client.
 *  telem  CCBTelemetryMessage *  The message to be delivered.
 * Output:
 *  return                 int    0 - OK.
 *                                1 - Error.
 */
static int tc_deliver_telemetry(CCBTelemetryClient *tc,
				CCBTelemetryMessage *telem)
{
  int no_callback = 0;   /* True if a callback is missing */
  int waserr = 0;        /* True if an error occurs */
  CCBTimeStamp ts;       /* The time-stamp of the message */
/*
 * Get a snapshot of the container of the callbacks while we have
 * exclusive access to it.
 */
  CCBTelemetryClientCallbacks cb = tc->cb;
/*
 * Get the timestamp of the message.
 */
  ts.mjd = telem->head.mjd;
  ts.sec = telem->head.sec;
  ts.ns = telem->head.ns;
/*
 * To avoid the risk of a callback recursively locking 'tc' by calling
 * another function in this module, temporarily relinquish exclusive
 * access to 'tc' while we call the manager's callback.
 */
  tc_unlock_client(tc);
/*
 * Locate the callback associated with the type of message that has been
 * provided.
 */
  switch(telem->head.base.type) {
  case CCB_MONITOR_MSG:           /* A monitor-data telemetry message */
    if(cb.monitor_msg_fn) {
      CCBMonitorMsg *msg = &telem->monitor;
      waserr = cb.monitor_msg_fn(tc->client, cb.monitor_msg_data, &ts,
				 msg->scan, msg->id, msg->data,
				 CCB_MAX_MONITOR);
    } else {
      no_callback = 1;
    };
    break;
  case CCB_INTEG_MSG:           /* An integrated-data telemetry message */
    if(cb.integ_msg_fn) {
      CCBIntegMsg *msg = &telem->integ;
      waserr = cb.integ_msg_fn(tc->client, cb.integ_msg_data, &ts,
			       msg->scan, msg->id, msg->data, CCB_MAX_INTEG);
    } else {
      no_callback = 1;
    };
    break;
  case CCB_LOG_MSG:           /* A log message */
    if(cb.log_msg_fn) {
      CCBLogMsg *msg = &telem->log;
      waserr = cb.log_msg_fn(tc->client, cb.log_msg_data, &ts, msg->msg,
			     msg->id, msg->level);
    } else {
      no_callback = 1;
    };
    break;
  case CCB_TELEM_PING_REPLY:  /* A reply to a ping command */
    waserr = ccb_report_ping_reply(tc->client, CCB_TELEM_LINK,
				   telem->ping.seq);
    break;
  default:
    ccb_log_client_msg(tc->client, CCB_ERROR, CCB_LOGID(148),
      "tc_deliver_telemetry: Can't deliver unknown type of message.\n");
    waserr = 1;
    break;
  };
/*
 * Reacquire exclusive access to 'tc'.
 */
  if(tc_lock_client(tc))
    return 1;
/*
 * If the message type was known by the above switch statement, but
 * the client process hadn't registered a callback for it, report
 * this here.
 */
  if(no_callback) {
    waserr = tc_error_while_locked(tc, CCB_ERROR, CCB_LOGID(149),
		"No callback has been provided to deliver \"%s\" messages.\n",
		ccb_net_obj_name(&ccb_telemetry_msg_table,
				 telem->head.base.type));
  };
  return waserr;
}

/*.......................................................................
 * Return the file descriptor of the current telemetry-link socket, if
 * any.
 *
 * Input:
 *  tc   CCBTelemetryClient *  The networking interface of the telemetry
 *                             client.
 * Output:
 *  return              int    The socket file descriptor, or -1 if
 *                             not currently connected to a CCB telemetry
 *                             server.
 */
int ccb_tc_socket(CCBTelemetryClient *tc)
{
  int fd;   /* The file descriptor of the socket */
/*
 * Check the arguments.
 */
  if(!tc)
    return -1;
/*
 * Acquire exclusive access to the tc object.
 */
  if(tc_lock_client(tc))
    return -1;
/*
 * Get the file descriptor of the socket.
 */
  fd = tc_socket_fd(tc);
/*
 * Relinquish exclusive access to the tc object.
 */
  tc_unlock_client(tc);
  return fd;
}

/*.......................................................................
 * Return the file-descriptor of a given CCBTelemetryClient object. Note
 * that this is an internal function which assumes that its caller has
 * acquired the guard mutex of the object.
 *
 * Input:
 *  tc   CCBTelemetryClient *  The networking interface of the telemetry
 *                             client.
 * Output:
 *  return              int    The socket file descriptor, or -1 if
 *                             not currently connected to a CCB telemetry
 *                             server.
 */
static int tc_socket_fd(CCBTelemetryClient *tc)
{
  return tc ? nrs_get_fd(tc->nrs) : -1;
}

/*.......................................................................
 * Initiate a connection to a CCB telemetry server.
 *
 * Input:
 *  tc  CCBTelemetryClient *  The client side of the CCB telemetry link.
 *  host        const char *  The address of the CCB server to contact.
 *                            When the 'nonblocking' argument is zero,
 *                            this can be either a numeric IP address or
 *                            a host name. When the 'nonblocking' argument
 *                            is non-zero, only numeric IP addresses are
 *                            accepted, since name-server lookups take
 *                            time.
 *  nonblocking        int    When this argument is 0, this function
 *                            doesn't return until the telemetry link has
 *                            either been established or rejected. When
 *                            non-zero, a non-blocking connect is
 *                            initiated to the CCB telemetry server, and
 *                            this function may then return before
 *                            this connection has completed. When this
 *                            happens, responding to the completion or
 *                            rejection of the connection is left to
 *                            a subsequent call to ccb_tc_send_telemetry()
 *                            and/or ccb_tc_read_telemetry() depending on
 *                            which type of I/O the completion elicits.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_tc_connect(CCBTelemetryClient *tc, const char *host, int nonblocking)
{
  int status=0;   /* The return status while holding the client lock */
/*
 * Check the arguments.
 */
  if(!tc || !host) {
    ccb_log_client_msg(tc ? tc->client : NULL, CCB_ERROR, CCB_LOGID(150),
	"ccb_tc_connect: NULL argument(s).\n");    
    return 1;
  };
/*
 * Get exclusive access to the tc object.
 */
  if(tc_lock_client(tc))
    return 1;
/*
 * Attempt to connect to the telemetry server.
 */
  status = tc_connect_link(tc, host, nonblocking);
/*
 * Relinquish exclusive access to 'tc'.
 */
  tc_unlock_client(tc);
  return status;
}

/*.......................................................................
 * Initiate a connection to a CCB telemetry server.
 *
 * Note that the caller must acquire exclusive access to 'tc before
 * calling this function.
 *
 * Input:
 *  tc  CCBTelemetryClient *  The client side of the CCB telemetry link.
 *  host        const char *  The address of the CCB server to contact.
 *                            When the 'nonblocking' argument is zero,
 *                            this can be either a numeric IP address or
 *                            a host name. When the 'nonblocking' argument
 *                            is non-zero, only numeric IP addresses are
 *                            accepted, since name-server lookups take
 *                            time.
 *  nonblocking        int    When this argument is 0, this function
 *                            doesn't return until the telemetry link has
 *                            either been established or rejected. When
 *                            non-zero, a non-blocking connect is
 *                            initiated to the CCB telemetry server, and
 *                            this function may then return before
 *                            this connection has completed. When this
 *                            happens, responding to the completion or
 *                            rejection of the connection is left to
 *                            a subsequent call to ccb_tc_send_telemetry()
 *                            and/or ccb_tc_read_telemetry() depending on
 *                            which type of I/O the completion elicits.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
static int tc_connect_link(CCBTelemetryClient *tc, const char *host,
			   int nonblocking)
{
  int fd;         /* The file descriptor of the new socket */
  int pending;    /* True if a non-blocking connection is pending completion */
/*
 * Terminate any existing telemetry connection.
 */
  tc_disconnect_link(tc);
/*
 * Mark the link as being in the process of being connected.
 */
  tc->link_state = TC_LINK_CONNECTING;
/*
 * If a blocking connect is being used, or a name-server lookup being
 * performed, tcp_connect() may block indefinitely, so for now
 * release exclusive access to the tc object, so that other threads
 * can queue messages to be sent when the connection eventually
 * completes.
 */
  tc_unlock_client(tc);
/*
 * Attempt to connect to the CCB telemetry-port on the specified host.
 */
  errno = 0;
  fd = tcp_connect(host, CCB_TELEMETRY_PORT, !nonblocking);
/*
 * Is a non-blocking connect still in progress?
 */
  pending = nonblocking && errno == EINPROGRESS;
/*
 * Reacquire exclusive access to the tc object. If this fails,
 * cancel the new connection.
 */
  if(tc_lock_client(tc)) {
    tc->link_state = TC_LINK_DOWN;
    if(fd >= 0)
      close(fd);
    return 1;
  };
/*
 * Connection failed?
 */
  if(fd < 0) {
    tc->link_state = TC_LINK_DOWN;
    tc_error_while_locked(tc, CCB_ERROR, CCB_LOGID(151),
	       "Couldn't connect to CCB telemetry server at %s (%s).\n",
	       host, strerror(errno));
    return 1;
  };
/*
 * Did another thread request a disconnect or another connection while
 * we didn't have exclusive access to 'tc'?
 */
  if(tc->link_state != TC_LINK_CONNECTING) {
    close(fd);
  } else {
/*
 * Attach the network input iterator to the new socket.
 */
    attach_NetReadStr(tc->nrs, fd);
/*
 * If the connection completed successfully, attempt to activate the
 * telemetry link.
 */
    if(!pending && tc_activate_link(tc)) {
      tc_disconnect_link(tc);
      return 1;
    };
  };
  return 0;
}

/*.......................................................................
 * Disconnect from a remote CCB telemetry server.
 *
 * Input:
 *  tc  CCBTelemetryClient *  The client side of the CCB telemetry link.
 */
void ccb_tc_disconnect(CCBTelemetryClient *tc)
{
  if(tc) {
/*
 * Get exclusive access to the tc object.
 */
    if(tc_lock_client(tc))
      return;
/*
 * Disconnect the link.
 */
    tc_disconnect_link(tc);
/*
 * Relinquish exclusive access to the tc object.
 */
    tc_unlock_client(tc);
  };
  return;
}

/*.......................................................................
 * Disconnect from a remote CCB telemetry server.
 *
 * Note that the caller must acquire exclusive access to 'tc' before
 * calling this fucntion.
 *
 * Input:
 *  tc  CCBTelemetryClient *  The client side of the CCB telemetry link.
 */
static void tc_disconnect_link(CCBTelemetryClient *tc)
{
/*
 * Check that the call isn't redundant.
 */
  if(tc->link_state != TC_LINK_DOWN &&
     tc->link_state != TC_LINK_DISCONNECTING) {
    int fd;  /* The file descriptor of the socket */
/*
 * Record the fact that the link is in the process of being disconnected.
 */
    tc->link_state = TC_LINK_DISCONNECTING;
/*
 * Get the file-descriptor of the telemetry socket.
 */
    fd = tc_socket_fd(tc);
/*
 * Dettach the socket from the I/O iterator.
 */
    detach_NetReadStr(tc->nrs);
/*
 * Close the telemetry socket.
 */
    if(fd >= 0)
      close(fd);
/*
 * Mark the telemetry connection as down.
 */
    tc->link_state = TC_LINK_DOWN;
  };
  return;
}

/*.......................................................................
 * When performing a non-blocking connect() on a socket, when select()
 * says that the socket has become either readable or writable (or
 * both), this function is called to ascertain whether the connection
 * succeded or not.
 *
 * Input:
 *  tc  CCBTelemetryClient *  The client side of the CCB telemetry link.
 *  sock               int    The socket to check, to see if a non-blocking
 *                            connection succeded or failed.
 * Output:
 *  return             int    1 - The connection completed successfully.
 *                            0 - The connection failed.
 */
static int tc_connection_completed(CCBTelemetryClient *tc, int sock)
{
  int status;                   /* The error status returned by getsockopt() */
  socklen_t status_len = sizeof(status); /* The size of 'status' */
/*
 * Query the socket's error flag. Note that whereas most implementations return
 * 0, and leave the error status in the 'status' argument, under Solaris
 * getsockopt() returns -1, and the error is left in errno.
 */
  if(getsockopt(sock, SOL_SOCKET, SO_ERROR, &status, &status_len) < 0)
    status = errno;
/*
 * Connection failed?
 */
  if(status) {
    tc_error_while_locked(tc, CCB_ERROR, CCB_LOGID(152),
	       "Couldn't connect to CCB telemetry server (%s).\n",
		strerror(status));
    return 0;
  };
/*
 * Activate the link.
 */
  if(tc_activate_link(tc))
    return 0; /* Failed */
/*
 * The connection succeded.
 */
  return 1;
}

/*.......................................................................
 * When a new connection is established the following function is called
 * to activate the link.
 *
 * Input:
 *  tc  CCBTelemetryClient *  The client side of the CCB telemetry link.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
static int tc_activate_link(CCBTelemetryClient *tc)
{
  int status;  /* The return value of a POSIX function */
/*
 * Mark the link as up.
 */
  tc->link_state = TC_LINK_UP;
/*
 * Tell any waiting readers that they can now procede to call read().
 */
  status = pthread_cond_broadcast(&tc->herald);
  if(status) {
    ccb_log_client_msg(tc->client, CCB_ERROR, CCB_LOGID(258),
		       "tc_activate_link: pthread_cond_broadcast -> %s\n",
		       strerror(status));
    return 1;
  };
  return 0;
}

/*.......................................................................
 * Acquire exclusive access to a given CCBTelemetryClient object.
 *
 * Input:
 *  tc  CCBTelemetryClient *  The client side of the CCB telemetry link.
 * Output:
 *  return             int    0 - Exclusive access granted.
 *                            1 - Couldn't get access.
 */
static int tc_lock_client(CCBTelemetryClient *tc)
{
  int status = pthread_mutex_lock(&tc->guard);
  if(status) {
    ccb_log_client_msg(tc->client, CCB_ERROR, CCB_LOGID(153),
		       "Couldn't get telemetry-client lock.\n");
    return 1;
  }
  return 0;
}

/*.......................................................................
 * Relinquish exclusive access to a given CCBTelemetryClient object.
 *
 * Input:
 *  tc  CCBTelemetryClient *  The client side of the CCB telemetry link.
 * Output:
 *  return             int    0 - Access relinquished.
 *                            1 - Error.
 */
static int tc_unlock_client(CCBTelemetryClient *tc)
{
  int status = pthread_mutex_unlock(&tc->guard);
/*
 * If unlocking fails we can't report the error via the manager's callback,
 * since we still have the mutex locked, so pass NULL instead of tc->client,
 * to tell ccb_log_client_msg() to write to stderr.
 */
  if(status) {
    ccb_log_client_msg(NULL, CCB_ERROR, CCB_LOGID(154),
		       "Couldn't release telemetry-client lock.\n");
    return 1;
  }
  return 0;
}

/*.......................................................................
 * This function is used to safely report errors when the caller is
 * holding the mutex guard of a CCBTelemetryClient object. The return value
 * is 1, which is the error-return values of most functions in this file,
 * so it can be used as the expression of return statements.
 *
 * Input:
 *  tc  CCBTelemetryClient *  The client side of the CCB telemetry link.
 *  level      CCBLogLevel    The significance of the message.
 *  id       unsigned long    The identifier of the calling statement.
 *  fmt         const char *  The printf-style format string of the
 *                            message.
 *  ...                       Any arguments referred to by the format.
 * Output:
 *  return             int    Always 1.
 */
static int tc_error_while_locked(CCBTelemetryClient *tc, CCBLogLevel level,
				 unsigned long id, const char *fmt, ...)
{
  va_list ap;               /* The encapsulated variable argument list */
/*
 * Relinquish exclusive access to 'tc' while reporting the error.
 * This allows the manager's callback function to be invoked without
 * fear that it might call a function in the library that attempted
 * to recursively acquire the guard mutex (which would hang the program).
 */
  tc_unlock_client(tc);
/*
 * Format the error message, then deliver it to the manager.
 */
  va_start(ap, fmt);
  ccb_vlog_client_msg(tc->client, level, id, fmt, ap);
  va_end(ap);
/*
 * Reacquire exclusive access to 'tc', since the caller is expecting
 * to still have it when we return.
 */
  (void) tc_lock_client(tc);
/*
 * Return the value that most functions in this file usually return
 * on error.
 */
  return 1;
}

/*.......................................................................
 * This is a vfprintf-style function that composes an error message and
 * then uses the manager's log-reporting callback to deliver it.
 *
 * Input:
 *  tc  CCBTelemetryClient *  The client side of the CCB telemetry link,
 *                            or NULL to write to stderr.
 *  level      CCBLogLevel    The significance of the message.
 *  id       unsigned long    The numeric identifier of the message. This must
 *                            be chosen to be unique within the CCB software,
 *                            and be the lowest so-far unused id number >= 0.
 *  fmt         const char *  This is a standard printf-style format string.
 *  ap             va_list    The encapsulated list of arguments refered to
 *                            by the format specifiers in the fmt[] argument.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_tc_vlog_client_msg(CCBTelemetryClient *tc, CCBLogLevel level,
			   unsigned long id, const char *fmt, va_list ap)
{
  int status;               /* The return value of ccb_vsnprintf() */
  int retval = 0;           /* The return value of this function */
  char buffer[CCB_MAX_LOG]; /* The buffer in which the message is formatted */
  CCBTimeStamp ts;          /* The time-stamp of the message */
  CCBLogMsgFn *fn;          /* The log-message callback. */
  void *data;               /* log_msg_fn() callback data. */
/*
 * Check the arguments.
 */
  if(!fmt) {
    fmt = "ccb_log_client_msg: Missing log-message format string.";
    level = CCB_ERROR;
    retval = 1;
  };
/*
 * Do we have a telemetry-client object?
 */
  if(tc) {
/*
 * Temporarily acquire exclusive access to 'tc' while we extract
 * information from it.
 */
    if(tc_lock_client(tc))
      return 1;
/*
 * Get the callback function used to deliver log messages.
 */
    fn = tc->cb.log_msg_fn;
    data = tc->cb.log_msg_data;
/*
 * Increment the count of the number of recursive calls to this function
 * that are currently taking place. This allows us to avoid an infinite
 * recursion if any of the functions that this function calls, themselves
 * call this function.
 */
    tc->log_recursion_count++;
/*
 * Relinquish exclusive access to 'tc'.
 */
    tc_unlock_client(tc);
  } else {
    fn = 0;
    data = NULL;
  };
/*
 * Compose the message.
 */
  status = ccb_vsnprintf(buffer, sizeof(buffer), fmt, ap);
/*
 * Error formatting message?
 */
  if(status < 0) {
    strcpy(buffer, "Error formatting log message.");
    level = CCB_ERROR;
    retval = 1;
/*
 * If the message contains any newline characters, terminate the string
 * at the first one found.
 */
  } else {
    char *cptr = strchr(buffer, '\n');
    if(cptr)
      *cptr = '\0';
  };
/*
 * Get the current date and time.
 */
  if(ccb_get_timestamp(&ts))
    ts.mjd = ts.sec = ts.ns = 0;
/*
 * If a tc object has been provided, and this function isn't being
 * called recursively, deliver the message to the manager. Otherwise
 * write it to stderr.
 */
  if(tc && fn && tc->log_recursion_count == 1) {
    retval = fn(tc->client, data, &ts, buffer, id, level);
  } else {
    unsigned h,m,s;
    ccb_hms_of_timestamp(ts, &h, &m, &s);
    (void) fprintf(stderr, "%lu:%02uh%02um%02us [id=%lu]: %s\n", ts.mjd, h, m,
		   s, id, buffer);
  };
/*
 * Decrement the recursion count of the function, now that we are leaving it.
 */
  if(tc) {
    if(tc_lock_client(tc))
      retval = 1;
    tc->log_recursion_count--;
    tc_unlock_client(tc);
  };
  return retval;
}

/*.......................................................................
 * Select between blocking and non-blocking telemetry-client socket I/O.
 *
 * Input:
 *  tc   CCBTelemetryClient *  The networking interface of the telemetry
 *                             client.
 *  on                  int    True to select non-blocking I/O.
 *                             False to select blocking I/O.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_tc_non_blocking_io(CCBTelemetryClient *tc, int on)
{
  int fd;   /* The file-descriptor of the current telemetry socket */
/*
 * Check the arguments.
 */
  if(!tc) {
    ccb_log_client_msg(tc ? tc->client : NULL, CCB_ERROR, CCB_LOGID(205),
	"ccb_tc_non_blocking_io: NULL argument(s).\n");
    return 1;
  };
/*
 * Get the file-descriptor of the telemetry socket.
 */
  fd = tc_socket_fd(tc);
/*
 * Establish the requested form of I/O.
 */
  if(fd >= 0 && tcp_set_blocking(fd, !on)) {
    ccb_log_client_msg(tc->client, CCB_ERROR, CCB_LOGID(206),
	"Error establishing %sblocking socket I/O.", on ? "non-":"");    
    return 1;
  };
  return 0;
}
