#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "ccbtelemetryserver.h"
#include "ccblogid.h"
#include "ccblogtrail.h"
#include "netbuf.h"
#include "ccbnetobj.h"
#include "ccbconstants.h"
#include "ccbqueue.h"
#include "ccbtelemetrymsg.h"

/*
 * CCBTelemetryServer objects manage the telemetry link of the CCB server.
 * Note that ccbtelemetry.h provides a CCBTelemetryServer typedef alias
 * for this structure.
 */
struct CCBTelemetryServer {
  CCBTelemServerStatus state; /* The operational status of the telemetry */
                              /*  server. */
  CCBServerLink *server;      /* The parent CCB server */
  NetSendStr *nss;            /* The network output stream */
  CCBLogTrail *lt;            /* Checksums of recently sent log messages */
  CCBMessageQueue *lq;        /* The queue of outgoing log messages */
  CCBMessageQueue *iq;        /* The queue of outgoing integration messages */
  CCBMessageQueue *mq;        /* The queue of outgoing monitor messages */
  CCBMessageQueue *tq;        /* The queue of outgoing ping replies */
  unsigned selected_streams;  /* The currently selected set of telemetry */
                              /*  streams (a union of CCBTelemetryStream */
                              /*  enumerators). */
  unsigned active_streams;    /* The union of selected_streams and the */
                              /*  completement of the set of any streams */
                              /*  that are disabled during a standby. */
  unsigned short monitor_period; /* The monitoring period */
  unsigned short monitor_gap; /* The number of integrations that have been */
                              /*  sent to the manager since the last monitor */
                              /*  packet was sent. */
};

static int ts_stage_message(CCBTelemetryServer *ts, CCBTelemetryHeader *header);
static int ts_stage_next_message(CCBTelemetryServer *ts);
static int ts_unqueue_next_message(CCBTelemetryServer *ts,
				   CCBTelemetryMessage *msg);
static int ts_queue_telemetry_message(CCBTelemetryServer *ts,
				      CCBTelemetryHeader *head);
static int ts_fill_telemetry_header(CCBTelemetryServer *ts,
				    CCBTelemetryHeader *head,
				    CCBTelemetryType type,
				    const CCBTimeStamp *t);

/*.......................................................................
 * Create a new CCBTelemetryServer object.
 *
 * Input:
 *  server           CCBServerLink *  The parent CCB server.
 * Output:
 *  return  CCBTelemetryServer *  The new object, or NULL on error.
 */
CCBTelemetryServer *new_CCBTelemetryServer(CCBServerLink *server)
{
  CCBTelemetryServer *ts;   /* The object to be returned */
  char errmsg[CCB_MAX_LOG]; /* An error message */
/*
 * Check the arguments.
 */
  if(!server) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(76),
		"new_CCBTelemetryServer: NULL argument(s).\n");
    return NULL;
  };
/*
 * Check that the table of telemetry message descriptions is valid.
 */
  if(ccb_check_net_table(&ccb_telemetry_msg_table, errmsg, sizeof(errmsg))) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(77),
		"Telemetry message-table error: %s.\n", errmsg);
    return NULL;
  };
/*
 * Allocate the container.
 */
  ts = malloc(sizeof(CCBTelemetryServer));
  if(!ts) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(78),
		"new_CCBTelemetryServer: Insufficient memory.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBTelemetryServer().
 */
  ts->state = CCB_TELEM_SERVER_DOWN;
  ts->server = server;
  ts->nss = NULL;
  ts->lt = NULL;
  ts->lq = NULL;
  ts->iq = NULL;
  ts->mq = NULL;
  ts->tq = NULL;
  ts->selected_streams = CCB_ALL_STREAMS;
  ts->active_streams = ts->selected_streams;
  ts->monitor_period = 1;
  ts->monitor_gap = 0;
/*
 * Allocate the output stream manager.
 */
  ts->nss = new_NetSendStr(-1,
		   ccb_net_max_obj_size(&ccb_telemetry_msg_table));
  if(!ts->nss) {
    ccb_log_server_msg(server, CCB_ERROR, CCB_LOGID(79),
	"new_CCBTelemetryServer: Insufficient memory for output stream.\n");
    return del_CCBTelemetryServer(ts);
  };
/*
 * Allocate a table of recently sent log messages.
 */
  ts->lt = new_CCBLogTrail(CCB_MAX_LOGID, CCB_MAX_LOG_VARIANTS);
  if(!ts->lt) {
    ccb_log_server_msg(server, CCB_ERROR, CCB_LOGID(80),
	"new_CCBTelemetryServer: Insufficient memory for log-trail.\n");
    return del_CCBTelemetryServer(ts);
  };
/*
 * Allocate the log-message queue.
 */
  ts->lq = new_CCBMessageQueue(sizeof(CCBLogMsg), CCB_LOG_QLEN, CCB_MQ_DRAIN);
  if(!ts->lq) {
    ccb_log_server_msg(server, CCB_ERROR, CCB_LOGID(81),
	"new_CCBTelemetryServer: Insufficient memory for log queue.\n");
    return del_CCBTelemetryServer(ts);
  };
/*
 * Allocate the integration-message queue.
 */
  ts->iq = new_CCBMessageQueue(sizeof(CCBIntegMsg), CCB_INTEG_QLEN,
			       CCB_MQ_DRAIN);
  if(!ts->iq) {
    ccb_log_server_msg(server, CCB_ERROR, CCB_LOGID(82),
      "new_CCBTelemetryServer: Insufficient memory for integration queue.\n");
    return del_CCBTelemetryServer(ts);
  };
/*
 * Allocate the monitor-message queue.
 */
  ts->mq = new_CCBMessageQueue(sizeof(CCBMonitorMsg), CCB_MONITOR_QLEN,
			       CCB_MQ_WRAP);
  if(!ts->mq) {
    ccb_log_server_msg(server, CCB_ERROR, CCB_LOGID(83),
	"new_CCBTelemetryServer: Insufficient memory for monitor queue.\n");
    return del_CCBTelemetryServer(ts);
  };
/*
 * Allocate the queue of ping messages.
 */
  ts->tq = new_CCBMessageQueue(sizeof(CCBTelemPingReply), CCB_TELEM_LINK_QLEN,
			       CCB_MQ_WRAP);
  if(!ts->tq) {
    ccb_log_server_msg(server, CCB_ERROR, CCB_LOGID(84),
	"new_CCBTelemetryServer: Insufficient memory for ping-reply queue.\n");
    return del_CCBTelemetryServer(ts);
  };
  return ts;
}

/*.......................................................................
 * Delete a CCBTelemetryServer object.
 *
 * Input:
 *  ts     CCBTelemetryServer *  The object to be deleted.
 * Output:
 *  return CCBTelemetryServer *  The deleted object (always NULL).
 */
CCBTelemetryServer *del_CCBTelemetryServer(CCBTelemetryServer *ts)
{
  if(ts) {
    ts->nss = del_NetSendStr(ts->nss); 
    ts->lt = del_CCBLogTrail(ts->lt);
    ts->lq = del_CCBMessageQueue(ts->lq);
    ts->iq = del_CCBMessageQueue(ts->iq);
    ts->mq = del_CCBMessageQueue(ts->mq);
    ts->tq = del_CCBMessageQueue(ts->tq);
    free(ts);
  };
  return NULL;
}

/*.......................................................................
 * Stage a telemetry message for subsequent transmission to the manager.
 *
 * Input:
 *  ts     CCBTelemetryServer *  The telemetry server.
 *  head   CCBTelemetryHeader *  The header of a particular
 *                               telemetry message to stage for
 *                               transmission.
 * Output:
 *  return                int    0 - OK.
 *                               1 - Error.
 */
static int ts_stage_message(CCBTelemetryServer *ts, CCBTelemetryHeader *head)
{
/*
 * Mark the server as idle until a new message has successfully staged
 * for transmission.
 */
  ts->state = CCB_TELEM_SERVER_IDLE;
/*
 * Encode the message in the network buffer.
 */
  if(ccb_obj_to_net(&ccb_telemetry_msg_table, nss_get_NetBuf(ts->nss),
		    &head->base)) {
    ccb_log_server_msg(ts->server, CCB_ERROR, CCB_LOGID(85),
	"ts_stage_message: ccb_obj_to_net() error (%s).\n", strerror(errno));
    return 1;
  };
/*
 * Mark the telemetry link as busy.
 */
  ts->state = CCB_TELEM_SERVER_BUSY;
  return 0;
}

/*.......................................................................
 * Starting with the highest priority queue, examine the various
 * output queues of active streams, searching for the next message to
 * be sent. If one is found, unqueue it and return its contents.
 *
 * Input:
 *  ts     CCBTelemetryServer *  The telemetry server.
 *  msg   CCBTelemetryMessage *  The message will be recorded in *msg.
 * Output:
 *  return                int    0 - OK.
 *                               1 - No message returned.
 */
static int ts_unqueue_next_message(CCBTelemetryServer *ts,
				   CCBTelemetryMessage *msg)
{
/*
 * Examining from the highest to the lowest priority queues, get the
 * next message to be staged for transmission.
 */
  if(ccb_mq_pop(ts->tq, msg, sizeof(CCBTelemPingReply)) ==
     sizeof(CCBTelemPingReply))
    return 0;
  if((ts->active_streams & CCB_LOG_STREAM) &&
     ccb_mq_pop(ts->lq, msg, sizeof(CCBLogMsg)) == sizeof(CCBLogMsg))
    return 0;
  if((ts->active_streams & CCB_INTEG_STREAM) &&
     ccb_mq_pop(ts->iq, msg, sizeof(CCBIntegMsg)) == sizeof(CCBIntegMsg))
    return 0;
  if((ts->active_streams & CCB_MONITOR_STREAM) &&
     ccb_mq_pop(ts->mq, msg, sizeof(CCBMonitorMsg)) == sizeof(CCBMonitorMsg))
    return 0;
/*
 * All of the queues of active streams are empty, so there is nothing
 * to be returned.
 */
  return 1;
}

/*.......................................................................
 * Set up the contents of the header of a telemetry message.
 *
 * Input:
 *  ts    CCBTelemetryServer *  The inquiring telemetry server.
 *  head  CCBTelemetryHeader *  The header to be initilized.
 *  type    CCBTelemetryType    The type of message.
 *  t     const CCBTimeStamp *  The time-stamp to record with the message,
 *                              or NULL to use the current time.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
static int ts_fill_telemetry_header(CCBTelemetryServer *ts,
				    CCBTelemetryHeader *head,
				    CCBTelemetryType type,
				    const CCBTimeStamp *t)
{
  CCBTimeStamp now;   /* The current time */
/*
 * If no time-stamp has been provided, substitute the current time.
 */
  if(!t) {
    if(ccb_get_timestamp(&now)) {
      ccb_log_server_msg(ts->server, CCB_ERROR, CCB_LOGID(86),
		  "ts_fill_telemetry_header: ccb_get_timestamp() failed.\n");
      return 1;
    };
    t = &now;
  };
/*
 * Fill in the header.
 */
  head->base.type = type;
  head->mjd = t->mjd;
  head->sec = t->sec;
  head->ns = t->ns;
  return 0;
}

/*.......................................................................
 * Queue a new log message.
 *
 * Input:
 *  ts    CCBTelemetryServer *  The queue of log messages.
 *  level        CCBLogLevel    The significance level of the new message.
 *  id         unsigned long    The numeric identifier of the originating
 *                              statement.
 *  text          const char *  The text to be queued.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
int ccb_ts_queue_log_msg(CCBTelemetryServer *ts, CCBLogLevel level,
			 unsigned long id, const char *text)
{
  CCBLogMsg msg;        /* The log message structure */
/*
 * Check the arguments.
 */
  if(!ts || !text || id >= CCB_MAX_LOGID) {
    if(!ts || !text)
      ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(87),
		  "ccb_ts_queue_log_msg: NULL argument(s).\n");
    else
      ccb_log_server_msg(ts->server, CCB_ERROR, CCB_LOGID(88),
	  "ccb_ts_queue_log_msg: Log-statement ID out of range.\n");
    return 1;
  };
/*
 * Reject redundant and excessive messages from the originating statement.
 */
  if(!ccb_lt_verify_msg(ts->lt, id, text))
    return 0;
/*
 * Copy the message details into a log-data communications structure.
 */
  if(ts_fill_telemetry_header(ts, &msg.head, CCB_LOG_MSG, NULL))
    return 1;
  msg.id = CCB_BASE_LOGID + id; /* Convert from CCB log-ID to Ygor log-ID */
  msg.level = level;
  strncpy(msg.msg, text, CCB_MAX_LOG);
  msg.msg[CCB_MAX_LOG-1] = '\0';
/*
 * If no message is currently staged for transmission, stage the new message.
 * Otherwise append it to the queue of outgoing log messages.
 */
  if(ts_queue_telemetry_message(ts, &msg.head))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a new integration message.
 *
 * Input:
 *  ts   CCBTelemetryServer *   The telemetry server.
 *  t    const CCBTimeStamp *   The time at which the integration was
 *                              recorded.
 *  scan      unsigned long     The index of the scan that originated
 *                              the integration.
 *  number    unsigned long     The sequential number of the integration
 *                              within the scan.
 *  values    unsigned long *   The integrated values.
 *  nvalues   unsigned long     The number of values in values[]. This
 *                              should be CCB_MAX_INTEG. If less than
 *                              this the remaining values will be set
 *                              to zero, and if more are passed, the
 *                              extra elements will be ignored.
 * Output:
 *  return              int     0 - OK.
 *                              1 - Error.
 */
int ccb_ts_queue_integ_msg(CCBTelemetryServer *ts, const CCBTimeStamp *t,
			   unsigned long scan, unsigned long number,
			   const unsigned long *values, unsigned nvalues)
{
  CCBIntegMsg msg;   /* The new message */
  int i;
/*
 * Check the arguments.
 */
  if(!ts || !t || !values) {
    ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(89),
		"ccb_ts_queue_integ_msg: NULL argument(s).\n");
    return 1;
  };
/*
 * Silently discard integration-data messages while the telemetry link is down.
 * This is necessary, because the scan numbers recorded in integration
 * messages won't make sense to the next manager that connects.
 */
  if(ts->state == CCB_TELEM_SERVER_DOWN)
    return 0;
/*
 * Compose the integration message.
 */
  if(ts_fill_telemetry_header(ts, &msg.head, CCB_INTEG_MSG, t))
    return 1;
  msg.scan = scan;
  msg.id = number;
  if(nvalues > CCB_MAX_INTEG)
    nvalues = CCB_MAX_INTEG;
  for(i=0; i<nvalues; i++)
    msg.data[i] = values[i];
  for(   ; i<CCB_MAX_INTEG; i++)
    msg.data[i] = 0;
/*
 * If no message is currently staged for transmission, stage the new message.
 * Otherwise append it to queue of outgoing integration messages.
 */
  if(ts_queue_telemetry_message(ts, &msg.head))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a new monitor message.
 *
 * Input:
 *  ts   CCBTelemetryServer *   The telemetry server.
 *  t    const CCBTimeStamp *   The time at which the monitor was
 *                              recorded.
 *  scan      unsigned long     The sequential number of the parent scan.
 *  number    unsigned long     The sequential number of the monitor
 *                              within the scan.
 *  values    unsigned long *   The integrated values.
 *  nvalues   unsigned long     The number of values in values[]. This
 *                              should be CCB_MAX_MONITOR. If less than
 *                              this, the remaining values will be set
 *                              to zero, and if more are passed, the
 *                              extra elements will be ignored.
 * Output:
 *  return              int     0 - OK.
 *                              1 - Error.
 */
int ccb_ts_queue_monitor_msg(CCBTelemetryServer *ts, const CCBTimeStamp *t,
			     unsigned long scan, unsigned long number,
			     const unsigned long *values, unsigned nvalues)
{
  CCBMonitorMsg msg;   /* The new message */
  int i;
/*
 * Check the arguments.
 */
  if(!ts || !t || !values) {
    ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(90),
		"ccb_ts_queue_monitor_msg: NULL argument(s).\n");
    return 1;
  };
/*
 * Silently discard monitor-data messages while the telemetry link is down.
 */
  if(ts->state == CCB_TELEM_SERVER_DOWN)
    return 0;
/*
 * Increment of the count of the number of samples that have been
 * reported since the last monitoring message was sent to the manager.
 */
  ts->monitor_gap++;
/*
 * If there hasn't been a large enough gap in the monitoring since the
 * last monitor-data message was sent, simply ignore this new packet.
 */
  if(ts->monitor_gap < ts->monitor_period)
    return 0;
/*
 * We are about to send a monitoring message, so reset the recorded
 * monitoring gap to zero.
 */
  ts->monitor_gap = 0;
/*
 * Compose the telemetry message.
 */
  if(ts_fill_telemetry_header(ts, &msg.head, CCB_MONITOR_MSG, t))
    return 1;
  msg.scan = scan;
  msg.id = number;
  if(nvalues > CCB_MAX_MONITOR)
    nvalues = CCB_MAX_MONITOR;
  for(i=0; i<nvalues; i++)
    msg.data[i] = values[i];
  for(   ; i<CCB_MAX_MONITOR; i++)
    msg.data[i] = 0;
/*
 * If no message is currently staged for transmission, stage the new message.
 * Otherwise append it to queue of outgoing monitor messages.
 */
  if(ts_queue_telemetry_message(ts, &msg.head))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a ping reply for subsequent transmission to the manager.
 *
 * Input:
 *  ts   CCBTelemetryServer *   The telemetry server.
 *  seq       unsigned long     The sequence number of the ping command.
 * Output:
 *  return              int     0 - OK.
 *                              1 - Error.
 */
int ccb_ts_queue_telem_ping_reply(CCBTelemetryServer *ts, unsigned long seq)
{
  CCBTelemPingReply msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!ts) {
    ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(91),
		"ccb_ts_queue_telem_ping_reply: NULL argument(s).\n");
    return 1;
  };
/*
 * Silently discard integration-data messages while the telemetry link is down.
 */
  if(ts->state == CCB_TELEM_SERVER_DOWN)
    return 0;
/*
 * Compose the telemetry message.
 */
  if(ts_fill_telemetry_header(ts, &msg.head, CCB_TELEM_PING_REPLY, NULL))
    return 1;
/*
 * Fill in teh body of the message.
 */
  msg.seq = seq;
/*
 * If no message is currently staged for transmission, stage the new message.
 * Otherwise append it to the queue of outgoing ping replies.
 */
  if(ts_queue_telemetry_message(ts, &msg.head))
    return 1;
  return 0;
}

/*.......................................................................
 * Attempt to send queued telemetry messages.
 *
 * Input:
 *  ts   CCBTelemetryServer *   The telemetry server.
 * Output:
 *  return              int     0 - OK.
 *                              1 - Error.
 */
int ccb_ts_send_telemetry(CCBTelemetryServer *ts)
{
/*
 * Check the arguments.
 */
  if(!ts) {
    ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(92),
		"ccb_ts_send_telemetry: NULL argument(s).\n");
    return 1;
  };
/*
 * Keep writing until either all messages have been sent, or the output
 * socket blocks.
 */
  while(ts->state == CCB_TELEM_SERVER_BUSY) {
/*
 * Attempt to write the latest message to the network.
 */
    switch(nss_send_msg(ts->nss)) {
    case NET_SEND_DATA:     /* Output temporarily blocked */
      return 0;
      break;
    case NET_SEND_DONE:     /* Message written, stage the next message */
/*
 * Mark the telemetry server as idle until we have a new message to be sent.
 */
      ts->state = CCB_TELEM_SERVER_IDLE; /* No more messages to send */
/*
 * Get a new message from the queue, if possible, and stage it for
 * transmission. Note that if a new message is staged, ts->state
 * will change to CCB_TELEM_SERVER_BUSY.
 */
      if(ts_stage_next_message(ts))
	return 1;
      break;
    case NET_SEND_CLOSED:
      ccb_log_server_msg(ts->server, CCB_ERROR, CCB_LOGID(93),
	  "The CCB server attempted to write to a closed telemetry socket.\n");
      return 1;
      break;
    case NET_SEND_ERROR:
    default:
      ccb_log_server_msg(ts->server, CCB_ERROR, CCB_LOGID(94),
    "The CCB server encountered an I/O error while sending telemetry data.\n");
      return 1;
      break;
    };
  };
/*
 * To get here, all queued messages must now have been sent.
 */
  return 0;
}

/*.......................................................................
 * Since no incoming data is expected over the telemetry connection,
 * if the caller indicates that the telemetry socket has become
 * readable, this either means that the remote end has closed the
 * connection, and thus that an end-of-file is waiting to be read,
 * or that somebody is trying to hack. In either case, report an error.
 *
 * Input:
 *  ts   CCBTelemetryServer *   The telemetry server.
 * Output:
 *  return              int     0 - OK.
 *                              1 - Error.
 */
int ccb_ts_read_telemetry(CCBTelemetryServer *ts)
{
  char byte;   /* A test byte */
/*
 * Check the arguments.
 */
  if(!ts) {
    ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(95),
		"ccb_ts_read_telemetry: NULL argument(s).\n");
    return 1;
  };
/*
 * Connection already closed?
 */
  if(ts->state == CCB_TELEM_SERVER_DOWN) {
    ccb_log_server_msg(ts->server, CCB_ERROR, CCB_LOGID(96),
	"Rejecting request to read from disconnected telemetry connection.\n");
    return 1;
  };
/*
 * Attempt to read a single byte from the telemetry link.
 */
  switch(read(nss_get_fd(ts->nss), &byte, 1)) {
  case 1:
    ccb_log_server_msg(ts->server, CCB_ERROR, CCB_LOGID(97),
        "The telemetry server reports the receipt of unexpected data.\n");
    break;
  case 0:
    ccb_log_server_msg(ts->server, CCB_ERROR, CCB_LOGID(98),
	"The manager appears to have closed the telemetry link.\n");
    break;
  default:
    ccb_log_server_msg(ts->server, CCB_ERROR, CCB_LOGID(99),
      "The CCB server got a read-error from the telemetry link.\n");
    break;
  };
/*
 * All of the above conditions are problematic.
 */
  return 1;
}

/*.......................................................................
 * Detach the telemetry server from the telemetry socket.
 *
 * Input:
 *  ts   CCBTelemetryServer *   The telemetry server.
 */
void ccb_ts_detach(CCBTelemetryServer *ts)
{
/*
 * Detach from the current telemetry socket.
 */
  detach_NetSendStr(ts->nss);
/*
 * Mark the server as down.
 */
  ts->state = CCB_TELEM_SERVER_DOWN;
  return;
}

/*.......................................................................
 * Attach the telemetry server to a new socket.
 *
 * Input:
 *  ts   CCBTelemetryServer *   The telemetry server.
 *  fd                  int     The file descriptor of the new telemetry
 *                              socket.
 */
int ccb_ts_attach(CCBTelemetryServer *ts, int fd)
{
/*
 * Check the arguments.
 */
  if(!ts || fd < 0) {
    if(!ts) {
      ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(100),
		  "ccb_ts_attach: NULL argument(s).\n");
    } else {
      ccb_log_server_msg(ts->server, CCB_ERROR, CCB_LOGID(101),
		  "ccb_ts_attach: Invalid socket.\n");
    };
    return 1;
  };
/*
 * Detach from any existing socket.
 */
  ccb_ts_detach(ts);
/*
 * Attach the output stream to the new socket.
 */
  attach_NetSendStr(ts->nss, fd);
/*
 * Mark the connection as idle.
 */
  ts->state = CCB_TELEM_SERVER_IDLE;
/*
 * Stage for transmission the highest priority of any messages that
 * remain in any of the queues.
 */
  if(ts_stage_next_message(ts))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a message for transmission to the manager. This involves
 * either adding it to a holding queue, or encoding it for
 * transmission, depending on whether or not the server is currently
 * busy transmitting a previous message.
 *
 * Input:
 *  ts    CCBTelemetryServer *  The telemetry server.
 *  head  CCBTelemetryHeader *  The header of a new message to queue.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
static int ts_queue_telemetry_message(CCBTelemetryServer *ts,
				      CCBTelemetryHeader *head)
{
/*
 * If enabled, queue the message according to its type.
 */
  switch(head->base.type) {
  case CCB_INTEG_MSG:
    if(ts->active_streams & CCB_INTEG_STREAM)
      (void) ccb_mq_push(ts->iq, head, sizeof(CCBIntegMsg));
    break;
  case CCB_MONITOR_MSG:
    if(ts->active_streams & CCB_MONITOR_STREAM)
      (void) ccb_mq_push(ts->mq, head, sizeof(CCBMonitorMsg));
    break;
  case CCB_LOG_MSG:
    if(ts->active_streams & CCB_LOG_STREAM)
      (void) ccb_mq_push(ts->lq, head, sizeof(CCBLogMsg));
    break;
  case CCB_TELEM_PING_REPLY:
    (void) ccb_mq_push(ts->tq, head, sizeof(CCBTelemPingReply));
    break;
  default:
    ccb_log_server_msg(ts->server, CCB_ERROR, CCB_LOGID(102),
	       "ccb_queue_telemetry_message: Unknown type of message.\n");
    return 1;
    break;
  };
/*
 * Stage the new message for transmission if the server is idle.
 */
  if(ts_stage_next_message(ts))
    return 1;
  return 0;
}

/*.......................................................................
 * If possible, stage the next of any queued messages for transmission.
 *
 * Input:
 *  ts       CCBTelemetryServer *  The telemetry server.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
static int ts_stage_next_message(CCBTelemetryServer *ts)
{
  CCBTelemetryMessage msg;  /* A dequeued message */
/*
 * Is the server ready to send a new message?
 */
  if(ts->state == CCB_TELEM_SERVER_IDLE) {
/*
 * Get the highest priority message from the queues of active streams
 * and prepare it for transmission.
 */
    if(ts_unqueue_next_message(ts, &msg) == 0 &&
       ts_stage_message(ts, &msg.head))
      return 1;
  };
  return 0;
}

/*.......................................................................
 * Query the current status of the telemetry server.
 *
 * Input:
 *  ts       CCBTelemetryServer *  The telemetry server.
 * Output:
 *  return CCBTelemServerStatus    The current status of the telemetry
 *                                 server.
 */
CCBTelemServerStatus ccb_ts_status(CCBTelemetryServer *ts)
{
  return ts ? ts->state : CCB_TELEM_SERVER_DOWN;
}

/*.......................................................................
 * Return non-zero if the integration queue is currently draining after
 * becoming full.
 *
 * Input:
 *  ts    CCBTelemetryServer *  The telemetry server.
 * Output:
 *  return               int    0 - Not draining.
 *                              1 - The queue of integration messages
 *                                  is currently closed while it
 *                                  dispatches existing messages.
 */
int ccb_ts_draining(CCBTelemetryServer *ts)
{
  return !ts || ccb_mq_draining(ts->iq);
}

/*.......................................................................
 * Select the set of telemetry streams that are to be sent to the manager.
 *
 * Input:
 *  ts    CCBTelemetryServer *  The telemetry server.
 *  streams         unsigned    The set of desired streams, expressed as
 *                              a bitmask union of CCBTelemetryStream
 *                              enumerators.
 * Output
 *  return               int    0 - OK.
 *                              1 - Error.
 */
int ccb_ts_select_streams(CCBTelemetryServer *ts, unsigned streams)
{
/*
 * Check the arguments.
 */
  if(!ts) {
    ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(221),
		"ccb_ts_select_streams: NULL argument(s).\n");
    return 1;
  };
/*
 * Install the new stream selection.
 */
  ts->selected_streams = streams;
  return 0;
}

/*.......................................................................
 * Specify the set of telemetry streams that are to be temporarily
 * disabled.
 *
 * Input:
 *  ts    CCBTelemetryServer *  The telemetry server.
 *  streams         unsigned    The set of streams to temporarily disable,
 *                              expressed as a bitmask union of
 *                              CCBTelemetryStream enumerators.
 * Output
 *  return               int    0 - OK.
 *                              1 - Error.
 */
int ccb_ts_mask_streams(CCBTelemetryServer *ts, unsigned streams)
{
/*
 * Check the arguments.
 */
  if(!ts) {
    ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(222),
		"ccb_ts_mask_streams: NULL argument(s).\n");
    return 1;
  };
/*
 * Combine the mask with the set of selected streams, to form the
 * set of streams that should actually be sent to the manager.
 */
  ts->active_streams = ts->selected_streams & ~streams;
  return 0;
}

/*.......................................................................
 * Set the period with which the server sub-samples the monitoring
 * packets.
 *
 * Input:
 *  ts   CCBTelemetryServer *  The networking interface of the CCB server.
 *  period   unsigned short    The monitoring period, expressed as a
 *                             multiple of the integration period.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_ts_monitor_period(CCBTelemetryServer *ts, unsigned short period)
{
/*
 * Check the arguments.
 */
  if(!ts) {
    ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(223),
		"ccb_ts_monitor_period: NULL argument(s).\n");
    return 1;
  };
/*
 * Record the new period.
 */
  ts->monitor_period = period;
  return 0;
}

/*.......................................................................
 * Revert the telemetry server to its startup state.
 *
 * Input:
 *  cs    CCBTelemetryServer *  The telemetry server.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
int ccb_ts_reset(CCBTelemetryServer *ts)
{
/*
 * Check the arguments.
 */
  if(!ts) {
    ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(231),
		"ccb_ts_reset: NULL argument(s).\n");
    return 1;
  };
/*
 * Clear legacy messages from all queues except the log queue.
 */
  ccb_mq_purge(ts->iq);
  ccb_mq_purge(ts->mq);
  ccb_mq_purge(ts->tq);
/*
 * Discard all log-message checksums, so that only messages that are
 * repeated during the new connection, are discarded.
 */
  ccb_lt_purge(ts->lt);
/*
 * Reset the telemetry stream-masks.
 */
  ts->selected_streams = CCB_ALL_STREAMS;
  ts->active_streams = ts->selected_streams;
  return 0;
}

/*.......................................................................
 * Set the interval at which historical log messages are forgotten.
 *
 * Input:
 *  ts   CCBTelemetryServer *  The networking interface of the CCB server.
 *  period   unsigned short    The interval at which the record of
 *                             historical log messages is purged (seconds).
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_ts_log_msg_lifetime(CCBTelemetryServer *ts, unsigned long period)
{
/*
 * Check the arguments.
 */
  if(!ts) {
    ccb_log_server_msg(ts ? ts->server : NULL, CCB_ERROR, CCB_LOGID(246),
		"ccb_ts_log_msg_lifetime: NULL argument(s).\n");
    return 1;
  };
/*
 * Install the new interval.
 */
  if(ccb_lt_lifetime(ts->lt, period))
    return 1;
  return 0;
}

