#ifndef ccbtelemetryserver_h
#define ccbtelemetryserver_h

#include "ccbserverlink.h"  /* Get the CCBServerLink typedef */

/*-------------------------------------------------------------------*
 * This header defines the communications interface of the telemetry *
 * link between the CCB server and the CCB manager.                  *
 *-------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*
 *              The server interface to the telemetry link               *
 *-----------------------------------------------------------------------*
 *
 * CCBTelemetryServer objects manage the telemetry link of the CCB server.
 */
typedef struct CCBTelemetryServer CCBTelemetryServer;

CCBTelemetryServer *new_CCBTelemetryServer(CCBServerLink *sl);
CCBTelemetryServer *del_CCBTelemetryServer(CCBTelemetryServer *telem);

/*
 * Queue a log message for subsequent transmission to the manager.
 */
int ccb_ts_queue_log_msg(CCBTelemetryServer *ts, CCBLogLevel level,
			 unsigned long id, const char *text);

/*
 * Queue an integration message for subsequent transmission to the manager.
 */
int ccb_ts_queue_integ_msg(CCBTelemetryServer *ts, const CCBTimeStamp *t,
			   unsigned long scan, unsigned long number,
			   const unsigned long *values, unsigned nvalues);

/*
 * Queue a monitor-data message for subsequent transmission to the manager.
 */
int ccb_ts_queue_monitor_msg(CCBTelemetryServer *ts, const CCBTimeStamp *t,
			     unsigned long scan, unsigned long number,
			     const unsigned long *values, unsigned nvalues);

/*
 * Queue a ping reply for subsequent transmission to the manager.
 */
int ccb_ts_queue_telem_ping_reply(CCBTelemetryServer *ts, unsigned long seq);

/*
 * Attempt to send all queued telemetry messages to the manager.
 */
int ccb_ts_send_telemetry(CCBTelemetryServer *ts);

/*
 * Respond to readability of the telemetry socket. This means that
 * either the manager has closed the connection, or that unexpected
 * data have arrived. The latter condition implies that the client
 * is a hacker masquerading as a manager process, so this function
 * always returns non-zero to indicate that a problematic event
 * has occurred.
 */
int ccb_ts_read_telemetry(CCBTelemetryServer *ts);

/*
 * Attach the telemetry server to a new socket (ie. a new manager connected).
 */
int ccb_ts_attach(CCBTelemetryServer *ts, int fd);

/*
 * Detach the telemetry server from the telemetry socket.
 */
void ccb_ts_detach(CCBTelemetryServer *ts);

/*
 * Enumerate the states that the telemetry server can be in.
 */
typedef enum {
  CCB_TELEM_SERVER_DOWN, /* Their is currently no telemetry link */
  CCB_TELEM_SERVER_IDLE, /* No messages are currently queued */
  CCB_TELEM_SERVER_BUSY  /* The server is in the midst of sending a message */
} CCBTelemServerStatus;

/*
 * Query the current status of the telemetry server.
 */
CCBTelemServerStatus ccb_ts_status(CCBTelemetryServer *ts);

/*
 * Return non-zero if the integration queue is currently draining after
 * becoming full.
 */
int ccb_ts_draining(CCBTelemetryServer *ts);

/*
 * Select the set of telemetry streams that are to be sent to the manager.
 */
int ccb_ts_select_streams(CCBTelemetryServer *ts, unsigned streams);

/*
 * Specify the set of telemetry streams that are to be temporarily
 * disabled.
 */
int ccb_ts_mask_streams(CCBTelemetryServer *ts, unsigned streams);

/*
 * Set the period with which the server sub-samples the monitoring
 * packets.
 */
int ccb_ts_monitor_period(CCBTelemetryServer *ts, unsigned short period);

/*
 * Reset the telemetry stream to its just-connected state.
 */
int ccb_ts_reset(CCBTelemetryServer *ts);

/*
 * Set the interval at which historical log messages are forgotten.
 */
int ccb_ts_log_msg_lifetime(CCBTelemetryServer *ts, unsigned long period);

#endif
