#ifndef ccbunsigned_h
#define ccbunsigned_h

/*
 * The following functions perform arithmetic on unsigned integers that
 * are a multiple of the size of the builtin (unsigned long) datatype.
 *
 * These integers are recorded as arrays of one or more unsigned long
 * integers, arranged such that the least significant element is the
 * element at the start of the array, and the most significant, the
 * element at the end of the array. The individual unsigned long
 * elements can be thought of as single digits within numbers of base
 * (sizeof(unsigned long) * CHAR_BIT). Passing a NULL pointer in place
 * of an array is intepretted identically to passing a zero valued
 * integer.
 */

/*
 * Set a given bit within 'a'.
 */
int ccb_ull_set_bit(unsigned long *a, unsigned na, unsigned bit);

/*
 * Clear a given bit within 'a'.
 */
void ccb_ull_clear_bit(unsigned long *a, unsigned na, unsigned bit);

/*
 * Return the value of a specified bit within 'a'.
 */
int ccb_ull_get_bit(const unsigned long *a, unsigned na, unsigned bit);

/*
 * Set an integer to zero.
 */
void ccb_ull_assign_zero(unsigned long *a, unsigned na);

/*
 * Perform an arithmetic shift left by a specified number of
 * bits. This is equivalent to multiplying by a power of two.
 */
void ccb_ull_shift_left(unsigned long *a, unsigned na, unsigned nbits);

/*
 * Perform an arithmetic shift right by a specified number of
 * bits. This is equivalent to dividing by a power of two.
 */
void ccb_ull_shift_right(unsigned long *a, unsigned na, unsigned nbits);

/*
 * Compare 'a' and 'b', returning -1 if a<b, 0 if a==b, or 1 if a>b.
 */
int ccb_ull_compare(const unsigned long *a, unsigned na,
		    const unsigned long *b, unsigned nb);

/*
 * Return true if a==0.
 */
int ccb_ull_is_zero(const unsigned long *a, unsigned na);

/*
 * Return the zero-relative index of the most significant non-zero bit
 * within 'a'.
 */
int ccb_ull_msb(const unsigned long *a, unsigned na);

/*
 * Record the value of a-b in dif.
 */
int ccb_ull_subtract(const unsigned long *a, unsigned na,
		     const unsigned long *b, unsigned nb,
		     unsigned long *dif, unsigned ndif);

/*
 * Record the value of a+b in sum.
 */
int ccb_ull_add(const unsigned long *a, unsigned na,
		const unsigned long *b, unsigned nb,
		unsigned long *sum, unsigned nsum);

/*
 * Record the value of a/b in quot, and the remainder of a/b in rem.
 */
int ccb_ull_divide(const unsigned long *a, unsigned na,
		   const unsigned long *b, unsigned nb,
		   unsigned long *quot, unsigned nquot,
		   unsigned long *rem, unsigned nrem);

/*
 * Record the value of a*b in prod.
 */
int ccb_ull_multiply(const unsigned long *a, unsigned na,
		     const unsigned long *b, unsigned nb,
		     unsigned long *prod, unsigned nprod,
		     unsigned long *work, unsigned nwork);

/*
 * Copy the contents of 'a' into 'b'.
 */
int ccb_ull_copy(const unsigned long *a, unsigned na,
		 unsigned long *b, unsigned nb);

/*
 * Read an extended unsigned integer from a string.
 */
int ccb_ull_input(const char *s, unsigned long *a, unsigned na,
		  unsigned long *work1, unsigned long nwork1,
		  unsigned long *work2, unsigned long nwork2,
		  char **endp);

/*
 * Write an extended unsigned integer to a string.
 */
int ccb_ull_output(char *s, size_t dim, const unsigned long *a, unsigned na,
		   unsigned long *work1, unsigned nwork1,
		   unsigned long *work2, unsigned nwork2);

#endif
